# Downloads documents from GPO's GovInfo.gov site, using sitemaps
# to efficiently determine what needs to be updated. See
# https://www.govinfo.gov/sitemaps for a list of collections.
# This service was formerly called "Fdsys."
#
# usc-run govinfo --collections=BILLS,STATUTE,...
# Download bill text (from the BILLS collection; there's also a bulk
# data BILLS collection but it has less in it), the Statues at Large,
# and other documents from GovInfo.gov's non-bulk-data collections.
#
# usc-run govinfo --bulkdata=BILLSTATUS,FR,...
# Download bill status, the Federal Register, and other documents
# from GovInfo.gov's bulk data collections. (The BILLS collection occurs
# both as a regular collection (bill text in multiple formats) and as
# a bulk data collection (just XML starting recently). Use --bulkdata=BILLS
# to get the bulk data collection.)
#
#   Options:
#
#   --years=2001[,2002,2004]
#   Comma-separated list of years to download from. Applies to collections
#   that are divided by year.
#
#   --congress=113[,114]
#   Comma-separated list of congresses to download from. Applies to bulk
#   data collections like BILLSTATUS that are grouped by Congress + Bill Type.
#
#   --type=hr[,s,sres]
#   Comma-separated list of bill types to download. Applies to bulk
#   data collections like BILLSTATUS that are grouped by Congress + Bill Type.
#
#   --extract=mods,pdf,text,xml,premis
#   Extract the MODS, PDF, text, XML, or PREMIS file associated
#   with each package from the downloaded package ZIP file.
#
#   --filter="regex"
#   Only stores files that match the regex. Regular collections
#   are matched against the package name (i.e. BILLS-113hconres66ih)
#   while bulk data items are matched against the their file path
#   (i.e. 113/1/hconres/BILLS-113hconres66ih.xml).
#
#   --cached|--force
#   Always/never use the cache.

from lxml import etree, html
import glob
import json
import re
import logging
import os
import os.path
import zipfile
from congress.tasks import utils

import rtyaml


# globals
GOVINFO_BASE_URL = "https://www.govinfo.gov/"
COLLECTION_BASE_URL = GOVINFO_BASE_URL + "app/details/"
BULKDATA_BASE_URL = GOVINFO_BASE_URL + "bulkdata/"
COLLECTION_SITEMAPINDEX_PATTERN = (
    GOVINFO_BASE_URL + "sitemap/{collection}_sitemap_index.xml"
)
BULKDATA_SITEMAPINDEX_PATTERN = (
    GOVINFO_BASE_URL + "sitemap/bulkdata/{collection}/sitemapindex.xml"
)
FDSYS_BILLSTATUS_FILENAME = "fdsys_billstatus.xml"

# for xpath
ns = {"x": "http://www.sitemaps.org/schemas/sitemap/0.9"}


# Main entry point


def run(options):
    # Process sitemaps.
    for collection in sorted(options.get("collections", "").split(",")):
        if collection != "":
            update_sitemap(
                COLLECTION_SITEMAPINDEX_PATTERN.format(collection=collection),
                None,
                [],
                options,
            )
    for collection in sorted(options.get("bulkdata", "").split(",")):
        if collection != "":
            update_sitemap(
                BULKDATA_SITEMAPINDEX_PATTERN.format(collection=collection),
                None,
                [],
                options,
            )


def update_sitemap(url, current_lastmod, how_we_got_here, options):
    """Updates the local cache of a sitemap file."""

    # Skip if the year or congress flags are set and this sitemap is
    # not for that year or Congress.
    if should_skip_sitemap(url, options):
        return []

    # For debugging, remember what URLs we are stepping through.
    how_we_got_here = how_we_got_here + [url]

    # Get the file paths to cache:
    # * the sitemap XML for future runs
    # * its <lastmod> date (which comes from the parent sitemap) so we know if we need to re-download it now
    # * the <lastmod> dates of the packages listed in this sitemap so we know if we need to re-download any package files
    cache_file = get_sitemap_cache_file(url)
    cache_file = os.path.join("govinfo/sitemap", cache_file, "sitemap.xml")
    lastmod_cache_file = cache_file.replace(".xml", "-lastmod.yaml")
    lastmod_cache_file = os.path.join(utils.cache_dir(), lastmod_cache_file)
    if not os.path.exists(lastmod_cache_file):
        lastmod_cache = {}
    else:
        with open(lastmod_cache_file) as f:
            lastmod_cache = rtyaml.load(f)

    try:
        return update_sitemap2(
            url, current_lastmod, how_we_got_here, options, lastmod_cache, cache_file
        )
    finally:
        # Write the updated last modified dates to disk so we know the next time whether
        # we need to fetch the files. If we didn't download anything, no need to write an
        # empty file.
        with utils.NoInterrupt():
            with open(lastmod_cache_file, "w") as f:
                rtyaml.dump(lastmod_cache, f)


def update_sitemap2(
    url, current_lastmod, how_we_got_here, options, lastmod_cache, cache_file
):
    # Return a list of files we downloaded.
    results = []

    # Download anew if the current_lastmod doesn't match the stored lastmod
    # in our cache, and if --cache is not specified. Or if --force is given.
    # If we're not downloading it, load it from disk because we still have
    # to process each sitemap to ensure we've downloaded all of the package
    # files the user wants.
    download = should_download_sitemap(
        lastmod_cache.get("lastmod"), current_lastmod, options
    )

    # Download, or just retreive from cache.
    if download:
        logging.warn("Downloading: %s" % url)
    body = utils.download(
        url, cache_file, utils.merge(options, {"force": download, "binary": True})
    )
    if not body:
        logging.error("Failed to download %s. Skipping." % url)
        return results

    # If we downloaded a new file, update the lastmod for our cache.
    if download and current_lastmod:
        lastmod_cache["lastmod"] = current_lastmod

    # Load the XML.
    try:
        sitemap = etree.fromstring(body)
    except etree.XMLSyntaxError as e:
        raise Exception("XML syntax error in %s: %s" % (url, str(e)))

    # Process the entries.
    if sitemap.tag == "{http://www.sitemaps.org/schemas/sitemap/0.9}sitemapindex":

        # This is a sitemap index. Process the sitemaps listed in this
        # sitemapindex recursively.
        for node in sitemap.xpath("x:sitemap", namespaces=ns):
            # Get URL and lastmod date of the sitemap.
            url = str(node.xpath("string(x:loc)", namespaces=ns))
            lastmod = str(node.xpath("string(x:lastmod)", namespaces=ns))
            sitemap_results = update_sitemap(url, lastmod, how_we_got_here, options)
            if sitemap_results is not None:
                results = results + sitemap_results

    elif sitemap.tag == "{http://www.sitemaps.org/schemas/sitemap/0.9}urlset":

        # This is a regular sitemap with content items listed.

        # Process the items.
        for node in sitemap.xpath("x:url", namespaces=ns):
            url = str(node.xpath("string(x:loc)", namespaces=ns))
            lastmod = str(node.xpath("string(x:lastmod)", namespaces=ns))

            m = re.match(COLLECTION_BASE_URL + r"([^-]+)-(.*)", url)
            if m:
                collection = m.group(1)
                package_name = m.group(2)
                if options.get("filter") and not re.search(
                    options["filter"], package_name
                ):
                    continue
                try:
                    mirror_results = mirror_package(
                        collection,
                        package_name,
                        lastmod,
                        lastmod_cache.setdefault("packages", {}),
                        options,
                    )
                except:
                    logging.exception(
                        "Error fetching package {} in collection {} from {}.".format(
                            package_name, collection, url
                        )
                    )
                    mirror_results = []
                results.extend(mirror_results)

            else:
                # This is a bulk data item. Extract components of the URL.
                m = re.match(BULKDATA_BASE_URL + r"([^/]+)/(.*)", url)
                if not m:
                    raise Exception(
                        "Unmatched bulk data file URL (%s) at %s."
                        % (url, "->".join(how_we_got_here))
                    )
                collection = m.group(1)
                item_path = m.group(2)
                if options.get("filter") and not re.search(
                    options["filter"], item_path
                ):
                    continue
                try:
                    mirror_results = mirror_bulkdata_file(
                        collection, url, item_path, lastmod, options
                    )
                except:
                    logging.exception(
                        "Error fetching file {} in collection {} from {}.".format(
                            item_path, collection, url
                        )
                    )
                    mirror_results = None
                if mirror_results is not None and len(mirror_results) > 0:
                    results = results + mirror_results

    else:
        raise Exception(
            "Unknown sitemap type (%s) at the root sitemap of %s." % (sitemap.tag, url)
        )

    return results


def should_skip_sitemap(url, options):
    # Don't skip sitemap indexes.
    m = re.match(re.escape(GOVINFO_BASE_URL) + r"sitemap/(\w+)_sitemap_index.xml", url)
    if m:
        return False
    m = re.match(
        re.escape(GOVINFO_BASE_URL) + r"sitemap/bulkdata/(\w+)/sitemapindex.xml", url
    )
    if m:
        return False

    year_filter = options.get("years", "").strip()
    congress_filter = options.get("congress", "").strip()
    type_filter = options.get("type", "").strip()

    # Regular collections are grouped by publication year.
    # Which years should we download? All if none is specified.
    m = re.match(re.escape(GOVINFO_BASE_URL) + r"sitemap/(\w+)_(\d+)_sitemap.xml", url)
    if m:
        year = m.group(2)
        if year_filter != "" and year not in year_filter.split(","):
            return True

    # Bulk data collections are grouped into subdirectories that can
    # represent years (as in the FR collection) or other types of groupings
    # like Congress + Bill Type for the BILLSTATUS collection.
    m = re.match(
        re.escape(GOVINFO_BASE_URL) + r"sitemap/bulkdata/(\w+)/(\d+)(.*)/sitemap.xml",
        url,
    )
    if m:
        numeric_grouping = m.group(2)
        sitemap_type_grouping = m.group(3)  # E.g. 'hr' or 's'
        if year_filter != "" and numeric_grouping not in year_filter.split(","):
            return True
        if congress_filter != "" and numeric_grouping not in congress_filter.split(","):
            return True
        if type_filter != "" and sitemap_type_grouping not in type_filter.split(","):
            return True

    return False


def get_sitemap_cache_file(url):
    # Where should we store the local cache of the sitemap XML and a file
    # that stores its <lastmod> date for when we last downloaded it? Returns
    # a path relative to the cache root.

    m = re.match(re.escape(GOVINFO_BASE_URL) + r"sitemap/(\w+)_sitemap_index.xml", url)
    if m:
        return m.group(1)

    m = re.match(re.escape(GOVINFO_BASE_URL) + r"sitemap/(\w+)_(\d+)_sitemap.xml", url)
    if m:
        return m.group(1) + "/" + m.group(2)

    m = re.match(
        re.escape(GOVINFO_BASE_URL) + r"sitemap/bulkdata/(\w+)/sitemapindex.xml", url
    )
    if m:
        return m.group(1) + "-bulkdata"

    m = re.match(
        re.escape(GOVINFO_BASE_URL) + r"sitemap/bulkdata/(\w+)/(.+)/sitemap.xml", url
    )
    if m:
        return m.group(1) + "-bulkdata/" + m.group(2)

    raise ValueError(url)


def should_download_sitemap(lastmod_cache, current_lastmod, options):
    # Download a sitemap or just read from our cache?

    if not current_lastmod:
        # No lastmod is known for this file (it's the root of a sitemap
        # tree - this is the first web request).
        return True

    elif options.get("force", False):
        # User requests downloading everything.
        return True

    elif options.get("cached", False):
        # User requests downloading nothing.
        return False

    else:
        # Download if the lastmod from the parent sitemap doesn't agree with
        # the lastmod stored on disk.
        return current_lastmod != lastmod_cache


# Downloading Packages


def mirror_package(collection, package_name, lastmod, lastmod_cache, options):
    """Create a local mirror of a GovInfo.gov package."""

    # Where should we store the file? Each collection has a different
    # file system layout (for BILLS, we put bill text along where the
    # bills scraper puts bills).
    path = get_output_path(collection, package_name, options)
    if not path:  # should skip
        return []

    # Go to the part of the lastmod_cache for this package.
    lastmod_cache = lastmod_cache.setdefault(package_name, {})
    lastmod_cache = lastmod_cache.setdefault("files", {})

    # Download the package ZIP file. We don't know what formats are available
    # until we download this file, and when we hit particular package files
    # we get 302 HTTP responses, which is uninformative about whether the
    # file is supposed to exist or not. But we can reliably download the ZIP
    # package.
    file_path = os.path.join(path, "package.zip")

    # If the file was supposedly downloaded before (i.e. lastmod_cache is
    # not empty) but it is missing, force a re-download by clearing the lastmod cache.
    if lastmod_cache and not os.path.exists(file_path):
        logging.error(
            "Missing: " + file_path + " (previously: " + repr(lastmod_cache) + ")"
        )
        lastmod_cache.clear()

    # Download the package ZIP file if it's updated.
    downloaded_files = []
    if mirror_package_zipfile(
        collection, package_name, file_path, lastmod, lastmod_cache, options
    ):
        downloaded_files.append(file_path)

    # Extract files from the package ZIP file depending on the --extract
    # command-line arguments. We do this even if the package ZIP file has
    # not changed because the --extract arguments might have changed and
    # the caller may want to extract files after having already gotten the
    # package ZIP file.
    try:
        extracted_files = extract_package_files(
            collection, package_name, file_path, lastmod_cache, options
        )
        downloaded_files.extend(extracted_files)
    except zipfile.BadZipfile as e:
        # Sometimes files don't download properly. If the ZIP file is
        # corrupt, log the error and delete the file.
        logging.error(str(e) + ". Deleting: " + file_path, exc_info=True)
        os.unlink(file_path)

    return downloaded_files


def mirror_package_zipfile(
    collection, package_name, file_path, lastmod, lastmod_cache, options
):
    # Do we already have this file updated?
    if lastmod_cache.get("package") == lastmod:
        if not options.get("force", False):
            return

    # With --cached, skip if the file is already downloaded.
    if os.path.exists(file_path) and options.get("cached", False):
        return

    # Download.
    file_url = GOVINFO_BASE_URL + "content/pkg/{}-{}.zip".format(
        collection, package_name
    )
    logging.warn("Downloading: " + file_path)
    data = utils.download(
        file_url,
        file_path,
        utils.merge(
            options,
            {
                "binary": True,
                "force": True,  # decision to cache was made above
                "to_cache": False,
                "needs_content": False,
            },
        ),
    )

    # Update the lastmod of the downloaded file.
    lastmod_cache["package"] = lastmod
    return True


def extract_package_files(
    collection, package_name, package_file, lastmod_cache, options
):
    # Extract files from the package ZIP file depending on the --extract
    # command-line argument. When extracting a file, mark the extracted
    # file's lastmod as the same as the package's lastmod.

    # Get the formats that the user wants to extract.
    extract_formats = set(
        format for format in options.get("extract", "").split(",") if format.strip()
    )

    # Make a mapping from file formats to a tuple of the filename found in the package ZIP
    # file and the filename that we will use to store the extracted format locally.
    format_paths = {
        "pdf": (
            "{collection}-{package_name}/pdf/{collection}-{package_name}.pdf",
            "document.pdf",
        ),
        "text": (
            "{collection}-{package_name}/html/{collection}-{package_name}.htm",
            "document.html",
        ),  # text wrapped in HTML!
        "xml": (
            "{collection}-{package_name}/xml/{collection}-{package_name}.xml",
            "document.xml",
        ),
        "mods": ("{collection}-{package_name}/mods.xml", "mods.xml"),
        "premis": ("{collection}-{package_name}/premis.xml", "premis.xml"),
    }

    # Extract only files if the package lastmod is newer than the file's lastmod.
    extract_formats = {
        format
        for format in extract_formats
        if lastmod_cache.get(format) is None
        or lastmod_cache[format] < lastmod_cache["package"]
    }

    # Don't even bother opening the ZIP file if there are no new files to extract.
    if not extract_formats:
        return []

    # Open the package ZIP file and try to extract files with names
    # we recognize.
    extracted_files = []
    with zipfile.ZipFile(package_file) as package:
        for format in extract_formats:
            if format not in format_paths:
                raise ValueError("invalid format: " + format)

            # Construct the expected path in the package ZIP file and the desired local filename.
            package_path, local_path = format_paths[format]
            package_path = package_path.format(
                collection=collection, package_name=package_name
            )
            local_path = os.path.join(os.path.dirname(package_file), local_path)

            # Extract it.
            try:
                with package.open(package_path) as f1:
                    with open(local_path, "wb") as f2:
                        f2.write(f1.read())
                        cb = options.get("callback", None)
                        if cb and callable(cb):
                            cb(local_path)
            except KeyError:
                # No file of this format is present in this package.
                continue
            finally:
                # Even if the file didn't exist, which is NOT an error condition
                # because not all packages have documents of all formats, update
                # the format's file's lastmod in our cache so that we don't try
                # to extract it again later, unless the package is updated.
                lastmod_cache[format] = lastmod_cache["package"]

            logging.warn("Extracted: " + local_path)
            extracted_files.append(local_path)

            # The "text" format files are put in an HTML container. Unwrap it into a .txt file.
            if format == "text":
                file_path_text = local_path.replace(".html", ".txt")
                logging.info("Unwrapping HTML to: " + file_path_text)
                with open(local_path) as f1:
                    with open(file_path_text, "wb") as f2:
                        f2.write(unwrap_text_in_html(f1.read()))
                extracted_files.append(file_path_text)

            if collection == "BILLS" and format == "mods":
                # When we download bill files, also create the text-versions/data.json file
                # which extracts commonly used components of the MODS XML, whenever we update
                # that MODS file.
                extract_bill_version_metadata(
                    package_name, os.path.dirname(package_file)
                )

    return extracted_files


def get_bill_id_for_package(package_name, with_version=True, restrict_to_congress=None):
    m = re.match(r"(\d+)([a-z]+)(\d+)([a-z][a-z0-9]*|)$", package_name)
    if not m:
        raise Exception("Unmatched bill document package name: " + package_name)
    congress, bill_type, bill_number, version_code = m.groups()

    if restrict_to_congress and int(congress) != int(restrict_to_congress):
        return None

    if not with_version:
        return ("%s%s-%s" % (bill_type, bill_number, congress), version_code)
    else:
        return "%s%s-%s-%s" % (bill_type, bill_number, congress, version_code)


def get_output_path(collection, package_name, options):
    # Where to store the document files?

    # The path will depend a bit on the collection.
    if collection == "BILLS":
        # Store with the other bill data ([congress]/bills/[billtype]/[billtype][billnumber]).
        bill_and_ver = get_bill_id_for_package(
            package_name,
            with_version=False,
            restrict_to_congress=options.get("congress"),
        )
        if not bill_and_ver:
            return None  # congress number does not match options["congress"]
        from congress.tasks.bills import output_for_bill

        bill_id, version_code = bill_and_ver
        return output_for_bill(
            bill_id, "text-versions/" + version_code, is_data_dot=False
        )

    elif collection == "CRPT":
        # Store committee reports in [congress]/crpt/[reporttype].
        m = re.match(r"(\d+)([hse]rpt)(\d+)$", package_name)
        if not m:
            raise ValueError(package_name)
        congress, report_type, report_number = m.groups()
        if options.get("congress") and congress != options.get("congress"):
            return None  # congress number does not match options["congress"]
        return "%s/%s/%s/%s/%s" % (
            utils.data_dir(),
            congress,
            collection.lower(),
            report_type,
            report_type + report_number,
        )

    else:
        # Store in govinfo/COLLECTION/PKGNAME.
        path = "%s/govinfo/%s/%s" % (utils.data_dir(), collection, package_name)
        return path


def unwrap_text_in_html(data):
    text_content = str(html.fromstring(data).text_content())
    return text_content.encode("utf8")


# Downloading bulk data files


def mirror_bulkdata_file(collection, url, item_path, lastmod, options):
    # Return a list of files we downloaded.
    results = []

    # Where should we store the file?
    path = "%s/govinfo/%s/%s" % (utils.data_dir(), collection, item_path)

    # For BILLSTATUS, store this along with where we store the rest of bill
    # status data.
    if collection == "BILLSTATUS":
        from congress.tasks.bills import output_for_bill

        bill_id, version_code = get_bill_id_for_package(
            os.path.splitext(os.path.basename(item_path.replace("BILLSTATUS-", "")))[0],
            with_version=False,
        )
        path = output_for_bill(bill_id, FDSYS_BILLSTATUS_FILENAME, is_data_dot=False)

    # Where should we store the lastmod found in the sitemap so that
    # we can tell later if the file has changed?
    lastmod_cache_file = os.path.splitext(path)[0] + "-lastmod.txt"

    # Do we already have this file up to date?
    if os.path.exists(lastmod_cache_file) and not options.get("force", False):
        if lastmod == utils.read(lastmod_cache_file):
            return

    # With --cached, skip if the file is already downloaded.
    if os.path.exists(path) and options.get("cached", False):
        return

    # Download.
    logging.warn("Downloading: " + path)
    data = utils.download(
        url,
        path,
        utils.merge(
            options,
            {
                "binary": True,
                "force": True,  # decision to cache was made above
                "to_cache": False,
            },
        ),
    )
    results.append(path)

    if not data:
        # Something failed.
        return

    # Write the current last modified date back to disk so we know the next time whether
    # we need to fetch the file again.
    utils.write(lastmod, lastmod_cache_file)

    return results


def extract_bill_version_metadata(package_name, text_path):
    bill_version_id = get_bill_id_for_package(package_name)

    bill_type, number, congress, version_code = utils.split_bill_version_id(
        bill_version_id
    )

    bill_version = {
        "bill_version_id": bill_version_id,
        "version_code": version_code,
        "urls": {},
    }

    mods_ns = {"mods": "http://www.loc.gov/mods/v3"}
    doc = etree.parse(os.path.join(text_path, "mods.xml"))
    locations = doc.xpath("//mods:location/mods:url", namespaces=mods_ns)

    for location in locations:
        label = location.attrib["displayLabel"]
        if "HTML" in label:
            format = "html"
        elif "PDF" in label:
            format = "pdf"
        elif "XML" in label:
            format = "xml"
        else:
            format = "unknown"
        bill_version["urls"][format] = location.text

    bill_version["issued_on"] = doc.xpath(
        "string(//mods:dateIssued)", namespaces=mods_ns
    )

    utils.write(
        json.dumps(
            bill_version, sort_keys=True, indent=2, default=utils.format_datetime
        ),
        output_for_bill_version(bill_version_id),
    )


def output_for_bill_version(bill_version_id):
    bill_type, number, congress, version_code = utils.split_bill_version_id(
        bill_version_id
    )
    return "%s/%s/bills/%s/%s%s/text-versions/%s/data.json" % (
        utils.data_dir(),
        congress,
        bill_type,
        bill_type,
        number,
        version_code,
    )



import unittest
from amendment_info import parse_amendment_actions
import datetime

# parsing various kinds of action text to extract metadata and establish state





class AmendmentActions(unittest.TestCase):

	def test_amendment_action_uses_amendments_plural(self):
		action = {
			'text': 'On agreeing to the Poe amendments (A009) Failed by recorded vote: 141 - 279 (Roll no. 164).',
			'type': 'action',
			'references': [],
			'acted_at': datetime.datetime(2005, 6, 17, 11, 16),
		}
		actions = [action]
		parse_amendment_actions(actions)
		action = actions[0]
		self.assertEqual(action['where'], 'h')
		self.assertEqual(action['type'], 'vote')
		self.assertEqual(action['vote_type'], 'vote')
		self.assertEqual(action['result'], 'fail')
		self.assertEqual(action['how'], 'roll')
		self.assertEqual(action['roll'], 164)

	def test_amendment_action_uses_as_modified_instead_of_as_amended(self):
		action = {
			'text': 'On agreeing to the Jackson-Lee (TX) amendment (A015) as modified Agreed to by recorded vote: 233 - 192 (Roll no. 412). (text as modified: CR H6290)',
			'type': 'action',
			'references': [],
			'acted_at': datetime.datetime(2005, 6, 17, 11, 16),
		}
		actions = [action]
		parse_amendment_actions(actions)
		action = actions[0]
		self.assertEqual(action['where'], 'h')
		self.assertEqual(action['type'], 'vote')
		self.assertEqual(action['vote_type'], 'vote')
		self.assertEqual(action['result'], 'pass')
		self.assertEqual(action['how'], 'roll')
		self.assertEqual(action['roll'], 412)

	def test_amendment_action_uses_capital_N_in_no(self):
		action = {
			'text': 'On agreeing to the Capps amendment (A028) Failed by recorded vote: 213 - 219 (Roll No. 129).',
			'type': 'action',
			'references': [],
			'acted_at': datetime.datetime(2005, 6, 17, 11, 16),
		}
		actions = [action]
		parse_amendment_actions(actions)
		action = actions[0]
		self.assertEqual(action['where'], 'h')
		self.assertEqual(action['type'], 'vote')
		self.assertEqual(action['vote_type'], 'vote')
		self.assertEqual(action['result'], 'fail')
		self.assertEqual(action['how'], 'roll')
		self.assertEqual(action['roll'], 129)




import json
import logging
import os
import re
import xmltodict

from congress.tasks import bill_info, amendment_info, govinfo, utils


def run(options):
    bill_id = options.get("bill_id", None)

    processor_func = process_bill

    if options.get("reparse_actions"):
        # Overrid default behavior.
        processor_func = reparse_actions

    if bill_id:
        to_fetch = bill_id.split(",")
    else:
        if options.get("matching_action_regex"):
            options["matching_action_regex"] = re.compile(
                options["matching_action_regex"]
            )

        to_fetch = get_bills_to_process(options)

        if not to_fetch:
            logging.warn("No bills changed.")
            return None

        limit = options.get("limit", None)
        if limit:
            to_fetch = to_fetch[: int(limit)]

    utils.process_set(to_fetch, processor_func, options)


def get_bills_to_process(options):
    # Return a generator over bill_ids that need to be processed.
    # Every time we process a bill we copy the fdsys_billstatus-lastmod.txt
    # file to data-fromfdsys-lastmod.txt, next to data.json. This way we
    # know when the GovInfo (formerly FDSys) XML file has changed.

    def get_data_path(*args):
        # Utility function to generate a part of the path
        # to data/{congress}/bills/{billtype}/{billtypenumber}
        # given as many path elements as are provided. args
        # is a list of zero or more of congress, billtype,
        # and billtypenumber (in order).
        args = list(args)
        if len(args) > 0:
            args.insert(1, "bills")
        return os.path.join(utils.data_dir(), *args)

    if not options.get("congress"):
        # Get a list of all congress directories on disk.
        # Filter out non-integer directory names, then sort on the
        # integer.
        def filter_ints(seq):
            for s in seq:
                try:
                    yield int(s)
                except:
                    # Not an integer.
                    continue

        congresses = sorted(filter_ints(os.listdir(get_data_path())))

        # If we're reprocessing actions, start with the 93rd Congress.
        # Before that we may have bill data from other sources that don't
        # conform to the usual action parsing logic.
        if options.get("reparse_actions"):
            congresses = filter(lambda c: c >= 93, congresses)
    else:
        congresses = sorted([int(c) for c in options["congress"].split(",")])

    # walk through congresses
    for congress in congresses:
        # turn this back into a string
        congress = str(congress)

        # walk through all bill types in that congress
        # (sort by bill type so that we proceed in a stable order each run)
        path = get_data_path(congress)
        if not os.path.exists(path):
            continue
        bill_types = [
            bill_type for bill_type in os.listdir(path) if not bill_type.startswith(".")
        ]

        for bill_type in sorted(bill_types):

            # walk through each bill in that congress and bill type
            # (sort by bill number so that we proceed in a normal order)
            path = get_data_path(congress, bill_type)
            if not os.path.exists(path):
                continue
            bills = [bill for bill in os.listdir(path) if not bill.startswith(".")]
            for bill_type_and_number in sorted(
                bills, key=lambda x: int(x.replace(bill_type, ""))
            ):

                bill_id = bill_type_and_number + "-" + congress

                if options.get("matching_action_regex"):
                    # Include bills that have an action that matches a regular expression.
                    fn = get_data_path(
                        congress, bill_type, bill_type_and_number, "data.json"
                    )
                    if os.path.exists(fn):
                        with open(fn) as f:
                            bill = json.load(f)
                            for action in bill["actions"]:
                                if action.get("text") and options[
                                    "matching_action_regex"
                                ].search(action["text"]):
                                    yield bill_id
                    continue  # don't check modification dates

                fn = get_data_path(
                    congress,
                    bill_type,
                    bill_type_and_number,
                    govinfo.FDSYS_BILLSTATUS_FILENAME,
                )
                if os.path.exists(fn):
                    # The GovInfo.gov bulk data file exists. Does our JSON data
                    # file need to be updated?
                    bulkfile_lastmod = utils.read(fn.replace(".xml", "-lastmod.txt"))
                    parse_lastmod = utils.read(
                        get_data_path(
                            congress,
                            bill_type,
                            bill_type_and_number,
                            "data-fromfdsys-lastmod.txt",
                        )
                    )
                    if bulkfile_lastmod != parse_lastmod or options.get("force"):
                        yield bill_id


def process_bill(bill_id, options):
    fdsys_xml_path = _path_to_billstatus_file(bill_id)
    logging.info("[%s] Processing %s..." % (bill_id, fdsys_xml_path))

    # Read FDSys bulk data file.
    xml_as_dict = read_fdsys_bulk_bill_status_file(fdsys_xml_path, bill_id)
    bill_data = form_bill_json_dict(xml_as_dict)
    if isinstance(bill_data, str):  # Non-error failure
        return {
            "ok": True,
            "saved": False,
            "reason": bill_data,
        }
    # Convert and write out data.json and data.xml.
    destination = os.path.dirname(fdsys_xml_path) + "/data.json"
    utils.write(
        json.dumps(bill_data, indent=2, sort_keys=True),
        destination,
        {"diff": options.get("diff")},
    )
    cb = options.get("callback", None)
    if cb and callable(cb):
        cb(destination)

    from congress.tasks.bill_info import create_govtrack_xml

    with open(os.path.dirname(fdsys_xml_path) + "/data.xml", "wb") as xml_file:
        xml_file.write(create_govtrack_xml(bill_data, options))

    if options.get("amendments", True):
        process_amendments(bill_id, xml_as_dict, options)

    # Mark this bulk data file as processed by saving its lastmod
    # file under a new path.
    utils.write(
        utils.read(_path_to_billstatus_file(bill_id).replace(".xml", "-lastmod.txt")),
        os.path.join(os.path.dirname(fdsys_xml_path), "data-fromfdsys-lastmod.txt"),
        {"diff": options.get("diff")},
    )

    return {
        "ok": True,
        "saved": True,
    }


def _path_to_billstatus_file(bill_id):
    return output_for_bill(
        bill_id, govinfo.FDSYS_BILLSTATUS_FILENAME, is_data_dot=False
    )


def read_fdsys_bulk_bill_status_file(fn, bill_id):
    fdsys_billstatus = utils.read(fn)
    return xmltodict.parse(
        fdsys_billstatus, force_list=("item", "amendment", "committeeReport", "link")
    )


def form_bill_json_dict(xml_as_dict):
    """
    Handles converting a government bulk XML file to legacy dictionary form.

    @param bill_id: id of the bill in format [type][number]-[congress] e.x. s934-113
    @type bill_id: str
    @return: dictionary of bill attributes
    @rtype: dict
    """

    from packaging.version import parse as parse_version

    try:
        schema_version = parse_version(xml_as_dict["billStatus"]["version"])
    except KeyError:  # no 'version' attribute is present before 2022-12-20
        schema_version = parse_version("1.0.0")

    bill_dict = xml_as_dict["billStatus"]["bill"]
    if schema_version >= parse_version("3.0.0"):
        bill_id = build_bill_id(
            bill_dict["type"].lower(), bill_dict["number"], bill_dict["congress"]
        )
    else:
        bill_id = build_bill_id(
            bill_dict["billType"].lower(),
            bill_dict["billNumber"],
            bill_dict["congress"],
        )
    titles = bill_info.titles_for(bill_dict["titles"]["item"])
    actions = bill_info.actions_for(
        bill_dict["actions"]["item"],
        bill_id,
        bill_info.current_title_for(titles, "official"),
    )
    status, status_date = bill_info.latest_status(
        actions, bill_dict.get("introducedDate", "")
    )

    if bill_dict.get("sponsors") is None and bill_dict["titles"]["item"][0][
        "title"
    ].startswith("Reserved "):
        logging.info(
            "[%s] Skipping reserved bill number with no sponsor (%s)"
            % (bill_id, bill_dict["titles"]["item"][0]["title"])
        )
        return bill_dict["titles"]["item"][0]["title"]  # becomes the 'reason'

    if bill_dict.get("sponsors") is None:
        by_request = False
    elif schema_version >= parse_version("3.0.0"):
        by_request = bill_dict["sponsors"]["item"][0]["isByRequest"] == "Y"
    else:
        by_request = bill_dict["sponsors"]["item"][0]["byRequestType"] is not None

    billCommittees = bill_dict.get("committees")
    if schema_version < parse_version("3.0.0"):
        billCommittees = (billCommittees or {})["billCommittees"]

    if schema_version >= parse_version("3.0.0"):
        legislativeSubjects = bill_dict.get("subjects", {}).get("legislativeSubjects")
    else:
        legislativeSubjects = bill_dict["subjects"]["billSubjects"][
            "legislativeSubjects"
        ]

    if schema_version >= parse_version("3.0.0"):
        billSummaries = bill_dict.get("summaries", {}).get("summary")
    else:
        billSummaries = bill_dict["summaries"]["billSummaries"]["item"]
    if billSummaries and not isinstance(billSummaries, list):
        billSummaries = [billSummaries]

    bill_data = {
        "bill_id": bill_id,
        "bill_type": bill_dict.get(
            "type" if schema_version >= parse_version("3.0.0") else "billType"
        ).lower(),
        "number": bill_dict.get(
            "number" if schema_version >= parse_version("3.0.0") else "billNumber"
        ),
        "congress": bill_dict.get("congress"),
        "url": billstatus_url_for(bill_id),
        "introduced_at": bill_dict.get("introducedDate", ""),
        "by_request": by_request,
        "sponsor": (
            bill_info.sponsor_for(bill_dict["sponsors"]["item"][0])
            if bill_dict.get("sponsors")
            else None
        ),
        "cosponsors": bill_info.cosponsors_for(bill_dict.get("cosponsors")),
        "actions": actions,
        "history": bill_info.history_from_actions(actions),
        "status": status,
        "status_at": status_date,
        "enacted_as": bill_info.slip_law_from(actions),
        "titles": titles,
        "official_title": bill_info.current_title_for(titles, "official"),
        "short_title": bill_info.current_title_for(titles, "short"),
        "popular_title": bill_info.current_title_for(titles, "popular"),
        "summary": bill_info.summary_for(billSummaries),
        # The top term's case has changed with the new bulk data. It's now in
        # Title Case. For backwards compatibility, the top term is run through
        # '.capitalize()' so it matches the old string. TODO: Remove one day?
        "subjects_top_term": (
            _fixup_top_term_case(bill_dict["policyArea"]["name"])
            if bill_dict.get("policyArea")
            else None
        ),
        "subjects": sorted(
            (
                [_fixup_top_term_case(bill_dict["policyArea"]["name"])]
                if bill_dict.get("policyArea")
                else []
            )
            + (
                [item["name"] for item in legislativeSubjects["item"]]
                if legislativeSubjects
                else []
            )
        ),
        "related_bills": bill_info.related_bills_for(bill_dict.get("relatedBills")),
        "committees": bill_info.committees_for(billCommittees),
        "amendments": bill_info.amendments_for(bill_dict.get("amendments")),
        "committee_reports": bill_info.committee_reports_for(
            bill_dict.get("committeeReports")
        ),
        "updated_at": bill_dict.get("updateDate", ""),
    }

    return bill_data


def _fixup_top_term_case(term):
    if term in ("Native Americans",):
        return term
    return term.capitalize()


def build_bill_id(bill_type, bill_number, congress):
    return "%s%s-%s" % (bill_type, bill_number, congress)


def billstatus_url_for(bill_id):
    bill_type, bill_number, congress = utils.split_bill_id(bill_id)
    return (
        govinfo.BULKDATA_BASE_URL
        + "BILLSTATUS/{0}/{1}/BILLSTATUS-{0}{1}{2}.xml".format(
            congress, bill_type, bill_number
        )
    )


def output_for_bill(bill_id, format, is_data_dot=True):
    bill_type, number, congress = utils.split_bill_id(bill_id)
    if is_data_dot:
        fn = "data.%s" % format
    else:
        fn = format
    return "%s/%s/bills/%s/%s%s/%s" % (
        utils.data_dir(),
        congress,
        bill_type,
        bill_type,
        number,
        fn,
    )


def process_amendments(bill_id, bill_amendments, options):
    amdt_list = bill_amendments["billStatus"]["bill"].get("amendments")
    if amdt_list is None:  # many bills don't have amendments
        return

    for amdt in amdt_list["amendment"]:
        amendment_info.process_amendment(amdt, bill_id, options)


def reparse_actions(bill_id, options):
    # Load an existing bill status JSON file.
    data_json_fn = output_for_bill(bill_id, "json")
    if not os.path.exists(data_json_fn):
        return {
            "ok": True,
            "saved": False,
            "reason": "no file",
        }
    source = utils.read(data_json_fn)
    bill_data = json.loads(source)

    # Munge data.
    from congress.tasks.bill_info import parse_bill_action

    title = bill_info.current_title_for(bill_data["titles"], "official")
    old_status = "INTRODUCED"
    for action in bill_data["actions"]:
        new_action, new_status = parse_bill_action(action, old_status, bill_id, title)
        if new_status:
            old_status = new_status
            action["status"] = new_status
        elif "status" in action:
            del action["status"]
        # clear out deleted keys
        for key in (
            "vote_type",
            "how",
            "where",
            "result",
            "roll",
            "suspension",
            "calendar",
            "under",
            "number",
            "committee",
            "pocket",
            "law",
            "congress",
        ):
            if key in action and key not in new_action:
                del action[key]
        action.update(new_action)

    status, status_date = bill_info.latest_status(
        bill_data["actions"], bill_data["introduced_at"]
    )
    bill_data["status"] = status
    bill_data["status_at"] = status_date

    wrote_any = False

    if options.get("diff"):
        confirmer = utils.show_diff_ask_ok
    else:
        # If no --diff is given, just check that
        # the content hasn't changed --- don't bother
        # writing out anything with identical content.
        def confirmer(source, revised, fn):
            return source != revised

    # Write new data.json file.
    revised = json.dumps(bill_data, indent=2, sort_keys=True)
    if confirmer(source, revised, data_json_fn):
        utils.write(revised, data_json_fn)
        wrote_any = True

    # Write new data.xml file.
    from congress.tasks.bill_info import create_govtrack_xml

    data_xml_fn = data_json_fn.replace(".json", ".xml")
    with open(data_xml_fn, "r") as xml_file:
        source = xml_file.read()
    revised = create_govtrack_xml(bill_data, options)
    if confirmer(source, revised.decode("utf8"), data_xml_fn):
        with open(data_xml_fn, "wb") as xml_file:
            xml_file.write(revised)
        wrote_any = True

    return {
        "ok": True,
        "saved": wrote_any,
        "reason": "no changes or changes skipped by user",
    }



from congress.tasks import utils
import os.path
import os
import re
import datetime
import json
import lxml.etree
import uuid
import logging
import mechanize
import zipfile
import io
import requests
import subprocess

from email.utils import parsedate
from time import mktime

# to get text files their is a new dependency; you need to have pdftotext.
# On Ubuntu, apt-get install poppler-utils. On OS X, install it via MacPorts
# with port install poppler, or via Homebrew with brew install poppler.

# options:
#
#    --chamber: "house" or "senate" to limit the parse to a single chamber
#    --load_by: Takes a range of House Event IDs. Give it the beginning and end IDs with a dash between, otherwise, it goes by the committee feeds.
#    --docs=False: Don't download (& convert to text) House committee documents


def run(options):
    # can limit it to one chamber
    chamber = options.get("chamber", None)
    if chamber and (chamber in ("house", "senate")):
        chambers = chamber
    else:
        chambers = ("house", "senate")

    load_by = options.get("load_by", None)

    # Load the committee metadata from the congress-legislators repository and make a
    # mapping from thomas_id and house_id to the committee dict. For each committee,
    # replace the subcommittees list with a dict from thomas_id to the subcommittee.
    utils.require_congress_legislators_repo()
    committees = {}
    for c in utils.yaml_load("congress-legislators/committees-current.yaml"):
        committees[c["thomas_id"]] = c
        if "house_committee_id" in c:
            committees[c["house_committee_id"] + "00"] = c
        c["subcommittees"] = dict(
            (s["thomas_id"], s) for s in c.get("subcommittees", [])
        )

    if "senate" in chambers:
        print("Fetching Senate meetings...")
        meetings = fetch_senate_committee_meetings(committees, options)
        print("Writing Senate meeting data to disk.")
        utils.write_json(meetings, output_for("senate"))

    if "house" in chambers:
        if load_by == None:
            print("Fetching House meetings...")
            meetings = fetch_house_committee_meetings(committees, options)
        else:
            print("Fetching House meetings by event_id...")
            meetings = fetch_meeting_from_event_id(committees, options, load_by)

        print("Writing House meeting data to disk.")
        utils.write_json(meetings, output_for("house"))

    # Write all meetings to a single file on disk.


# TODO: if these have unique IDs, maybe worth storing a file per-meeting.
def output_for(chamber):
    return utils.data_dir() + "/committee_meetings_%s.json" % chamber


# Parse the Senate committee meeting XML feed for meetings.
# To aid users of the data, attempt to assign GUIDs to meetings.
def fetch_senate_committee_meetings(committees, options):
    # Load any existing meetings file so we can recycle any GUIDs.
    existing_meetings = []
    output_file = output_for("senate")
    if os.path.exists(output_file):
        existing_meetings = json.load(open(output_file))

    options = dict(options)  # clone
    options["binary"] = True  #
    options["force"] = True

    meetings = []

    dom = lxml.etree.fromstring(
        utils.download(
            "https://www.senate.gov/general/committee_schedules/hearings.xml",
            "committee_schedule/senate.xml",
            options,
        )
    )

    for node in dom.xpath("meeting"):
        committee_id = str(node.xpath("string(cmte_code)"))
        if committee_id.strip() == "":
            continue  # "No committee hearings scheduled" placeholder
        occurs_at = str(node.xpath("string(date)"))
        room = str(node.xpath("string(room)"))
        topic = str(node.xpath("string(matter)"))

        occurs_at = datetime.datetime.strptime(occurs_at, "%d-%b-%Y %I:%M %p")
        topic = re.sub(r"\s+", " ", topic).strip()

        # Validate committee code.
        try:
            committee_code, subcommittee_code = re.match(
                r"(\D+)(\d+)$", committee_id
            ).groups()
            if committee_code not in committees:
                raise ValueError(committee_code)
            if subcommittee_code == "00":
                subcommittee_code = None
            if (
                subcommittee_code
                and subcommittee_code not in committees[committee_code]["subcommittees"]
            ):
                raise ValueError(subcommittee_code)
        except:
            print("Invalid committee code", committee_id)
            continue

        # See if this meeting already exists. If so, take its GUID.
        # Assume meetings are the same if they are for the same committee/subcommittee and
        # at the same time.
        for mtg in existing_meetings:
            if (
                mtg["committee"] == committee_code
                and mtg.get("subcommittee", None) == subcommittee_code
                and mtg["occurs_at"] == occurs_at.isoformat()
            ):
                if options.get("debug", False):
                    print("[%s] Reusing gUID." % mtg["guid"])
                guid = mtg["guid"]
                break
        else:
            # Not found, so create a new ID.
            # TODO: Can we make this a human-readable ID?
            guid = str(uuid.uuid4())

        # Scrape the topic text for mentions of bill numbers.
        congress = utils.congress_from_legislative_year(
            utils.current_legislative_year(occurs_at)
        )
        bills = []
        bill_number_re = re.compile(
            r"(hr|s|hconres|sconres|hjres|sjres|hres|sres)\s?(\d+)", re.I
        )
        for bill_match in bill_number_re.findall(topic.replace(".", "")):
            bills.append(bill_match[0].lower() + bill_match[1] + "-" + str(congress))

        # Create the meeting event.
        if options.get("debug", False):
            print(
                "[senate][%s][%s] Found meeting in room %s at %s."
                % (committee_code, subcommittee_code, room, occurs_at.isoformat())
            )

        meetings.append(
            {
                "chamber": "senate",
                "congress": congress,
                "guid": guid,
                "committee": committee_code,
                "subcommittee": subcommittee_code,
                "occurs_at": occurs_at.isoformat(),
                "room": room,
                "topic": topic,
                "bill_ids": bills,
            }
        )

    print("[senate] Found %i meetings." % len(meetings))
    return meetings


# House


# Scrape docs.house.gov for meetings.
# To aid users of the data, assign GUIDs to meetings piggy-backing off of the provided EventID.
def fetch_house_committee_meetings(committees, options):
    # Load any existing meetings file so we can recycle any GUIDs.
    existing_meetings = []
    output_file = output_for("house")
    if os.path.exists(output_file):
        existing_meetings = json.load(open(output_file))

    opts = dict(options)
    opts["binary"] = True
    opts["force"] = True

    meetings = []
    seen_meetings = set()

    # Scrape the committee listing page for a list of committees with scrapable events.
    committee_html = utils.download(
        "http://docs.house.gov/Committee/Committees.aspx",
        "committee_schedule/house_overview.html",
        opts,
    )
    for cmte in re.findall(r'<option value="(....)">', committee_html.decode("utf-8")):
        if cmte not in committees:
            logging.error("Invalid committee code: " + cmte)
            continue

        # Download the feed for this committee.
        logging.info("Fetching events for committee " + cmte)
        html = utils.download(
            "http://docs.house.gov/Committee/RSS.ashx?Code=%s" % cmte,
            "committee_schedule/house_%s.xml" % cmte,
            opts,
        )

        # It's not really valid?
        html = html.replace(
            b"&nbsp;", b" "
        )  # who likes nbsp's? convert to spaces. but otherwise, entity is not recognized.
        # print(html)
        # Parse and loop through the meetings listed in the committee feed.
        dom = lxml.etree.fromstring(html)

        # original start to loop
        for mtg in dom.xpath("channel/item"):

            eventurl = str(mtg.xpath("string(link)"))
            event_id = re.search(r"EventID=(\d+)$", eventurl)
            if not event_id:
                continue  # weird empty event showed up
            event_id = event_id.group(1)
            pubDate = datetime.datetime.fromtimestamp(
                mktime(parsedate(mtg.xpath("string(pubDate)")))
            )
            # skip old records of meetings, some of which just give error pages
            if pubDate < (datetime.datetime.now() - datetime.timedelta(days=60)):
                continue

            # Events can appear in multiple committee feeds if it is a joint meeting.
            if event_id in seen_meetings:
                logging.info("Duplicated multi-committee event: " + event_id)
                continue
            seen_meetings.add(event_id)

            # this loads the xml from the page and sends the xml to parse_house_committee_meeting
            load_xml_from_page(
                eventurl, options, existing_meetings, committees, event_id, meetings
            )
            # if bad zipfile
            if load_xml_from_page == False:
                continue

    print("[house] Found %i meetings." % len(meetings))
    return meetings


## load House meeting sequentially from event_id
def fetch_meeting_from_event_id(committees, options, load_id):
    existing_meetings = []
    output_file = output_for("house")
    if os.path.exists(output_file):
        existing_meetings = json.load(open(output_file))

    opts = dict(options)
    opts["binary"] = True
    opts["force"] = True

    meetings = []
    ids = load_id.split("-")
    current_id = int(ids[0])
    end_id = int(ids[1])

    while current_id <= end_id:
        event_id = str(current_id)
        event_url = (
            "http://docs.house.gov/Committee/Calendar/ByEvent.aspx?EventID=" + event_id
        )
        load_xml_from_page(
            event_url, options, existing_meetings, committees, event_id, meetings
        )
        # bad zipfile
        if load_xml_from_page == False:
            continue
        current_id += 1

    print("[house] Found %i meetings." % len(meetings))
    return meetings


def load_xml_from_page(
    eventurl, options, existing_meetings, committees, event_id, meetings
):
    # Load the HTML page for the event and use the mechanize library to
    # submit the form that gets the meeting XML. TODO Simplify this when
    # the House makes the XML available at an actual URL.

    logging.info(eventurl)
    package_info = extract_meeting_package(eventurl, event_id, options)
    if package_info == False:
        return False
    witnesses = package_info["witnesses"]
    uploaded_documents = package_info["uploaded_documents"]
    dom = package_info["dom"]

    # Parse the XML.
    try:
        meeting = parse_house_committee_meeting(
            event_id,
            dom,
            existing_meetings,
            committees,
            options,
            witnesses,
            uploaded_documents,
        )
        if meeting != None:  # an active meeting record
            meetings.append(meeting)

    except Exception as e:
        logging.error("Error parsing " + eventurl, exc_info=e)


# look for witnesses and documents in the house meeting package
def extract_meeting_package(eventurl, event_id, options):
    br = mechanize.Browser()
    # open committee event page
    br.open(eventurl)

    br.select_form(nr=0)

    # mechanize parser failed to find these fields
    br.form.new_control("hidden", "__EVENTTARGET", {})
    br.form.new_control("hidden", "__EVENTARGUMENT", {})
    br.form.set_all_readonly(False)

    # set field values
    if options.get("docs", True):
        # When we want documents, download the whole ZIP package.
        br["__EVENTTARGET"] = "ctl00$MainContent$LinkButtonDownloadMtgPackage"
    else:
        # Otherwise, just download the metadata XML.
        br["__EVENTTARGET"] = "ctl00$MainContent$LinkButtonDownloadMtgXML"
    br["__EVENTARGUMENT"] = ""

    # get the info
    request = br.submit()

    # when just downloading the metadata XML, return the DOM and no other info
    if not options.get("docs", True):
        try:
            dom = lxml.etree.fromstring(request.read())
        except lxml.etree.XMLSyntaxError as e:
            print(event_id, e)
            return False
        return {"witnesses": None, "uploaded_documents": [], "dom": dom}

    ## read zipfile
    try:
        request_bytes = io.BytesIO(request.read())
        package = zipfile.ZipFile(request_bytes)
    except:
        message = "Problem downloading zipfile: %s" % (event_id)
        print(message)
        return False

    # save documents in meeting package
    uploaded_documents = save_documents(package, event_id)
    witnesses = None
    # find meeting and witness xml
    for name in package.namelist():
        if ".xml" in name:
            if "WList" in name:
                bytes = package.read(name)
                witness_tree = lxml.etree.fromstring(bytes)
                witness_info = parse_witness_list(
                    witness_tree, uploaded_documents, event_id
                )
                witnesses = witness_info["hearing_witness_info"]
            else:
                bytes = package.read(name)
                dom = lxml.etree.fromstring(bytes)

    # it will return none if there is no witness list in the file
    return {
        "witnesses": witnesses,
        "uploaded_documents": uploaded_documents,
        "dom": dom,
    }


# parse xml for urls to testimony and witness information
def parse_witness_list(witness_tree, uploaded_documents, event_id):
    hearing_id = witness_tree.xpath("//@meeting-id")[0]
    hearing_witness_info = []
    # basic witness information
    for witness in witness_tree.xpath("panel/witness"):
        record = {"house_event_id": hearing_id}
        record["first_name"] = witness.xpath("string(firstname)")
        if record["first_name"] == "":
            record["first_name"] = None
        record["middle_name"] = witness.xpath("string(middlename)")
        if record["middle_name"] == "":
            record["middle_name"] = None
        record["last_name"] = witness.xpath("string(lastname)")
        if record["last_name"] == "":
            record["last_name"] = None
        record["honorific"] = witness.xpath("string(honorific)")
        if record["honorific"] == "":
            record["honorific"] = None
        record["position"] = witness.xpath("string(position)")
        if record["position"] == "":
            record["position"] = None
        record["organization"] = witness.xpath("string(organization)")
        if record["organization"] == "":
            record["organization"] = None
        record["witness_type"] = witness.xpath("string(witness-type)")
        if record["witness_type"] == "":
            record["witness_type"] = None
        record["documents"] = []
        # documents related to that witness
        for doc in witness.xpath("witness-documents/witness-document"):
            document = {}
            published_on = doc.xpath("string(@publish-date)")
            try:
                document["published_on"] = datetime.datetime.strptime(
                    published_on, "%Y-%m-%dT%H:%M:%S.%f"
                )
            except:
                document["published_on"] = datetime.datetime.strptime(
                    published_on, "%Y-%m-%dT%H:%M:%S"
                )

            document["description"] = doc.xpath("string(description)")
            if document["description"] == "":
                document["description"] = None

            doc_type = doc.xpath("string(type)")
            if doc_type == "":
                document["type"] = None
                document["type_name"] = None
            else:
                document["type"] = doc_type
                types = {
                    "CV": "Committee vote",
                    "WS": "Witness statement",
                    "WT": "Witness truth statement",
                    "WB": "Witness biography",
                    "CR": "Committee report",
                    "BR": "Bill",
                    "FA": "Floor amendment",
                    "CA": "Committee amendment",
                    "HT": "Transcript",
                    "WD": "Witness document",
                }
                # "SD": "" I don't know this one, the SD category covers a lot
                if doc_type in types:
                    document["type_name"] = types[doc_type]
                else:
                    document["type_name"] = None

            urls = []
            for files in doc.xpath("files/file"):
                url = files.xpath("string(@doc-url)")
                splinter = url.split("/")
                doc_name = splinter[-1]
                if doc_name not in uploaded_documents:
                    file_found = save_file(url, event_id)
                else:
                    file_found = True
                urls.append({"url": url, "file_found": file_found})

            document["urls"] = urls
            record["documents"].append(document)
        hearing_witness_info.append(record)
    return {"hearing_witness_info": hearing_witness_info}


# Grab a House meeting out of the DOM for the XML feed.
def parse_house_committee_meeting(
    event_id, dom, existing_meetings, committees, options, witnesses, uploaded_documents
):
    try:
        congress = int(dom.xpath("//@congress-num")[0])
        occurs_at = (
            dom.xpath("string(meeting-details/meeting-date/calendar-date)")
            + " "
            + dom.xpath("string(meeting-details/meeting-date/start-time)")
        )
        occurs_at = datetime.datetime.strptime(occurs_at, "%Y-%m-%d %H:%M:%S")
    except:
        raise ValueError(
            "Invalid meeting data (probably server error) in %s." % event_id
        )

    current_status = str(dom.xpath("string(current-status)"))
    if current_status not in ("S", "R"):
        # If status is "P" (postponed and not yet rescheduled) or "C" (cancelled),
        # don't include in output.
        return

    topic = dom.xpath("string(meeting-details/meeting-title)")

    committee_names = []
    for com in dom.xpath("meeting-details/committees"):
        comte = com.xpath("string(committee-name)")
        if comte != None:
            committee_names.append(com.xpath("string(committee-name)"))
    for scom in dom.xpath("meeting-details/subcommittees"):
        scomte = scom.xpath("string(committee-name)")
        if scomte != None:
            committee_names.append(scom.xpath("string(committee-name)"))

    room = None
    for n in dom.xpath("meeting-details/meeting-location/capitol-complex"):
        room = n.xpath("string(building)") + " " + n.xpath("string(room)")

    bills = []
    for bill_id in dom.xpath(
        "meeting-documents/meeting-document[@type='BR']/legis-num"
    ):
        # validating bill ids
        bill_id = house_bill_id_formatter(bill_id.text, congress)
        if bill_id != None:
            bills.append(bill_id)

    meeting_type = dom.xpath("//@meeting-type")[0]

    # Meeting documents include legislation, reports, etc.
    meeting_documents = []
    for doc in dom.xpath("//meeting-document"):
        document = {}
        published_on = doc.xpath("string(@publish-date)")
        try:
            document["published_on"] = datetime.datetime.strptime(
                published_on, "%Y-%m-%dT%H:%M:%S.%f"
            )
        except:
            document["published_on"] = datetime.datetime.strptime(
                published_on, "%Y-%m-%dT%H:%M:%S"
            )
        document["description"] = doc.xpath("string(description)")
        if document["description"] == "":
            document["description"] = None

        doc_type = doc.xpath("string(filename-metadata/doc-type)")
        if doc_type == "":
            document["type"] = None
            document["type_name"] = None
        else:
            document["type"] = doc_type
            types = {
                "CV": "Committee vote",
                "WS": "Witness statement",
                "WT": "Witness truth statement",
                "WB": "Witness biography",
                "CR": "Committee report",
                "BR": "Bill",
                "FA": "Floor amendment",
                "CA": "Committee amendment",
                "HT": "Transcript",
                "WD": "Witness document",
            }
            # "SD": "" I don't know this one, the SD category covers a lot
            if doc_type in types:
                document["type_name"] = types[doc_type]
            else:
                document["type_name"] = None

        document["bioguide_id"] = doc.xpath("string(filename-metadata/bioguideID)")
        if document["bioguide_id"] == "":
            document["bioguide_id"] = None

        document["amendmendment_number"] = doc.xpath(
            "string(filename-metadata/amdt-num)"
        )

        bill_id = doc.xpath("string(filename-metadata/legis-num)")
        document["bill_id"] = house_bill_id_formatter(bill_id, congress)

        document["version_code"] = doc.xpath("string(filename-metadata/legis-stage)")
        if document["version_code"] == "":
            document["version_code"] = None

        urls = []
        for u in doc.xpath("files/file"):
            url = u.xpath("string(@doc-url)")
            splinter = url.split("/")
            doc_name = splinter[-1]
            file_found = False
            if doc_name in uploaded_documents:
                file_found = True
            elif options.get("docs", True):
                file_found = save_file(url, event_id)
            urls.append({"url": url, "file_found": file_found})

        if len(urls) > 0:
            document["urls"] = urls
        meeting_documents.append(document)

    # Repeat the event for each listed committee or subcommittee, since our
    # data model supports only a single committee/subcommittee ID per event.

    orgs = []
    for c in dom.xpath("meeting-details/committees/committee-name"):
        if c.get("id") not in committees:
            raise ValueError("Invalid committee ID: " + c.get("id"))
        orgs.append((committees[c.get("id")]["thomas_id"], None))
    for sc in dom.xpath("meeting-details/subcommittees/committee-name"):
        if sc.get("id")[0:2] + "00" not in committees:
            raise ValueError("Invalid committee ID: " + sc.get("id"))
        c = committees[sc.get("id")[0:2] + "00"]
        if sc.get("id")[2:] not in c["subcommittees"]:
            logging.error("Invalid subcommittee code: " + sc.get("id"))
            continue
        orgs.append((c["thomas_id"], sc.get("id")[2:]))

    for committee_code, subcommittee_code in orgs:
        # See if this meeting already exists. If so, take its GUID.
        # Assume meetings are the same if they are for the same event ID and committee/subcommittee.
        for mtg in existing_meetings:
            if (
                mtg["house_event_id"] == int(event_id)
                and mtg.get("committee", None) == committee_code
                and mtg.get("subcommittee", None) == subcommittee_code
            ):
                guid = mtg["guid"]
                break
        else:
            # Not found, so create a new ID.
            # TODO: when does this happen?
            guid = str(uuid.uuid4())

        url = (
            "http://docs.house.gov/Committee/Calendar/ByEvent.aspx?EventID=" + event_id
        )

        # return the parsed meeting
        if options.get("debug", False):
            print(
                "[house][%s][%s] Found meeting in room %s at %s"
                % (committee_code, subcommittee_code, room, occurs_at.isoformat())
            )

        results = {
            "chamber": "house",
            "congress": congress,
            "guid": guid,
            "committee": committee_code,
            "committee_names": committee_names,
            "subcommittee": subcommittee_code,
            "occurs_at": occurs_at.isoformat(),
            "room": room,
            "topic": topic,
            "bill_ids": bills,
            "house_meeting_type": meeting_type,
            "house_event_id": int(event_id),
            "url": url,
        }

        # witness information and documents are only added if there was a result
        if witnesses != None:
            results["witnesses"] = witnesses
        if len(meeting_documents) > 0:
            results["meeting_documents"] = meeting_documents

        return results


# saves documents to disk, called in extract_meeting_package
def save_documents(package, event_id):
    uploaded_documents = []

    # find/create directory
    folder = str(int(event_id) / 100)
    output_dir = utils.data_dir() + "/committee/meetings/house/%s/%s" % (
        folder,
        event_id,
    )
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # loop through package and save documents
    for name in package.namelist():
        # for documents that are not xml
        if ".xml" not in name:
            try:
                bytes = package.read(name)
            except:
                print("Did not save to disk: file %s" % (name))
                continue
            file_name = "%s/%s" % (output_dir, name)

            # save document
            logging.info("saved " + file_name)
            with open(file_name, "wb") as document_file:
                document_file.write(bytes)
            # try to make a text version
            text_doc = text_from_pdf(file_name)
            if text_doc != None:
                uploaded_documents.append(text_doc)
            uploaded_documents.append(name)

    return uploaded_documents


# Code from IG scraper project
# uses pdftotext to get text out of PDFs,
# then writes it and returns the /data-relative path.
def text_from_pdf(pdf_path):
    try:
        subprocess.Popen(
            ["pdftotext", "-v"],
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ).communicate()
    except:
        logging.warn("Not processing PDF. Have you installed pdftotext?")
        return None

    real_pdf_path = pdf_path
    real_text_path = pdf_path.replace(".pdf", ".txt")

    try:
        subprocess.check_call(
            'pdftotext -layout "%s" "%s"' % (real_pdf_path, real_text_path), shell=True
        )
    except subprocess.CalledProcessError as exc:
        logging.warn("Error extracting text for %s\n" % (real_text_path))
        return None

    if os.path.exists(real_text_path):
        file_name = os.path.basename(real_text_path)
        return file_name
    else:
        logging.warn("Text not extracted to %s" % real_text_path)
        return None


# this is for files mentioned in the xml that do not appear in the meeting packet
def save_file(url, event_id):
    # not saving xml but I cold be convinced otherwise
    if ".xml" in url:
        return False

    r = requests.get(url, stream=True)

    if r.status_code == requests.codes.ok:
        # find or create directory
        folder = str(int(event_id) / 100)
        output_dir = utils.data_dir() + "/committee/meetings/house/%s/%s" % (
            folder,
            event_id,
        )
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        # get file name
        splinter = url.split("/")
        name = splinter[-1]
        file_name = "%s/%s" % (output_dir, name)
        # try to save

        try:
            logging.info("saved " + url + " to " + file_name)
            with open(file_name, "wb") as document_file:
                document_file.write(r.content)
            if ".pdf" in file_name:
                text_doc = text_from_pdf(file_name)
            return True
        except:
            print("Failed to save- %s" % (url))
            return False
    else:
        logging.info("failed to fetch: " + url)
        return False


def house_bill_id_formatter(bill_id, congress):
    # make sure there is a number
    if bill_id == None or bill_id == "":
        return None
    else:
        bill_id = bill_id.strip()
        digit = False
        alpha = False
        for char in bill_id:
            if char.isdigit():
                digit = True
            if char.isalpha():
                alpha = True

        if digit == False:
            return None
        # look for missing hr, though this risks mislabeling continuing and joint resolutions
        if alpha == False:
            bill_id = "hr" + bill_id
        else:
            bill_id = bill_id.replace(".", "").replace(" ", "").lower()

        bill_id = bill_id + "-" + str(congress)
        return bill_id



# Import the Adler & Wilkerson Congressional Bills Project
# data, which covers bills (but not resolutions) in the
# 80th through 92nd Congress,

import csv
import zipfile
import datetime

from congress.tasks import utils


def run(options):
    # Download the TSV file.
    cache_zip_path = "adler-wilkerson-bills.zip"
    utils.download(
        "http://congressionalbills.org/billfiles/bills80-92.zip",
        cache_zip_path,
        utils.merge(options, {'binary': True, 'needs_content': False}))

    # Unzip in memory and process the records.
    zfile = zipfile.ZipFile(utils.cache_dir() + "/" + cache_zip_path)
    csvreader = csv.DictReader(zfile.open("bills80-92.txt"), delimiter="\t")
    for record in csvreader:
        rec = process_bill(record)

        import pprint
        pprint.pprint(rec)


def process_bill(record):
    # Basic info
    congress = int(record["Cong"])
    bill_type = record["BillType"].lower()  # "HR" or "S" only
    if bill_type not in ('hr', 's'):
        raise ValueError(bill_type)
    number = int(record["BillNum"])

    def binary(value):
        if value == 'NULL':
            return None
        return value == '1'

    def nullydate(value):
        if value == 'NULL':
            return None
        raise ValueError(value)  # never occurs -- there are no dates in the dataset!

    # Last status?
    status = "INTRODUCED"
    status_at = nullydate(record['IntrDate'])
    if record['ReportH'] == '1' or record['ReportS'] == '1':
        status = "REPORTED"
    if record['PassH'] == '1' and record['PassS'] == '1':
        status = "PASSED:BILL"
    elif record['PassH'] == '1':
        status = "PASS_OVER:HOUSE"
    elif record['PassS'] == '1':
        status = "PASS_OVER:SENATE"
    if record['PLaw'] == '1':
        if record['Veto'] == '1':
            status = 'ENACTED:VETO_OVERRIDE'
        else:
            status = 'ENACTED:SIGNED'  # could also have been a 10-day rule
        status_at = nullydate(record['PLawDate'])
    else:
        if record['Veto'] == '1':
            status = 'PROV_KILL:VETO'

    # Form data structure
    return {
        'bill_id': "%s%d-%d" % (bill_type, number, congress),
        'bill_type': bill_type,
        'number': number,
        'congress': congress,

        'introduced_at': nullydate(record['IntrDate']),
        'sponsor': int(record['PooleID']) if record['PooleID'] != 'NULL' else None,
        #'cosponsors': ,

        #'actions': ,
        'history': {
            'house_passage_result': "pass" if record['PassH'] == '1' else None,
            'senate_passage_result': "pass" if record['PassS'] == '1' else None,
            'enacted': record['PLaw'] == '1',
            'enacted_at': nullydate(record['PLawDate']),
        },
        'status': status,
        'status_at': status_at,
        'enacted_as': {
            'law_type': "public",
            'congress': congress,
            'number': int(record['PLawNum']),
        } if record['PLaw'] == '1' else None,  # private laws?

        #'titles': ,
        'official_title': record['Title'],
        #'short_title': ,
        #'popular_title': ,

        #'summary': ,
        'subjects_top_term': int(record['Major']),
        'subjects': [int(record['Minor'])],

        #'related_bills': ,
        #'committees': ,
        #'amendments': ,

        # special fields
        'by_request': binary(record['ByReq']),
        'commemerative': binary(record['Commem']),
        'num_cosponsors': int(record['Cosponsr']) if record['Cosponsr'] != 'NULL' else None,
        'private': binary(record['Private']),

        # meta-metadata
        'updated_at': datetime.datetime.now(),
    }



import re
import logging
import datetime
import time
import json
from lxml import etree

from congress.tasks import utils

from congress.tasks.bill_info import sponsor_for as sponsor_for_bill, action_for


def process_amendment(amdt_data, bill_id, options):
    amdt = build_amendment_json_dict(amdt_data, options)
    path = output_for_amdt(amdt["amendment_id"], "json")

    logging.info("[%s] Saving %s to %s..." % (bill_id, amdt["amendment_id"], path))

    # output JSON - so easy!
    utils.write(
        json.dumps(amdt, sort_keys=True, indent=2, default=utils.format_datetime), path
    )

    with open(output_for_amdt(amdt["amendment_id"], "xml"), "wb") as xml_file:
        xml_file.write(create_govtrack_xml(amdt, options))


def build_amendment_json_dict(amdt_dict, options):
    # good set of tests for each situation:
    # samdt712-113 - amendment to bill
    # samdt112-113 - amendment to amendment on bill
    # samdt4904-111 - amendment to treaty
    # samdt4922-111 - amendment to amendment to treaty

    amendment_id = build_amendment_id(
        amdt_dict["type"].lower(), amdt_dict["number"], amdt_dict["congress"]
    )

    amends_bill = amends_bill_for(amdt_dict.get("amendedBill"))  # almost always present
    amends_treaty = None  # amends_treaty_for(amdt_dict) # the bulk data does not provide amendments to treaties (THOMAS did)
    amends_amendment = amends_amendment_for(
        amdt_dict.get("amendedAmendment")
    )  # sometimes present
    if not amends_bill and not amends_treaty:
        raise Exception("Choked finding out what bill or treaty the amendment amends.")

    actions = actions_for(amdt_dict["actions"]["actions"])

    amdt = {
        "amendment_id": amendment_id,
        "amendment_type": amdt_dict["type"].lower(),
        "chamber": amdt_dict["type"][0].lower(),
        "number": int(amdt_dict["number"]),
        "congress": amdt_dict["congress"],
        "amends_bill": amends_bill,
        "amends_treaty": amends_treaty,
        "amends_amendment": amends_amendment,
        "sponsor": sponsor_for(
            amdt_dict["sponsors"]["item"][0], amdt_dict["type"].lower()
        ),
        "purpose": amdt_dict["purpose"][0]
        if type(amdt_dict.get("purpose")) is list
        else amdt_dict.get("purpose"),
        "introduced_at": amdt_dict["submittedDate"][:10],
        "actions": actions,
        "updated_at": amdt_dict["updateDate"][0]
        if type(amdt_dict["updateDate"]) is list
        else amdt_dict["updateDate"],
    }

    # duplicate attributes creates lists when parsed, this block deduplicates
    if "description" in amdt_dict:
        amdt["description"] = amdt_dict.get("description")
        if type(amdt_dict.get("description")) is list:
            amdt["description"] = amdt["description"][0]

    if amdt_dict['type'][0].lower() == 's' and "proposedDate" in amdt_dict:
        amdt['proposed_at'] = amdt_dict['proposedDate']

    # needs to come *after* the setting of introduced_at
    amdt["status"], amdt["status_at"] = amendment_status_for(amdt)

    return amdt


def create_govtrack_xml(amdt, options):
    govtrack_type_codes = {
        "hr": "h",
        "s": "s",
        "hres": "hr",
        "sres": "sr",
        "hjres": "hj",
        "sjres": "sj",
        "hconres": "hc",
        "sconres": "sc",
    }
    root = etree.Element("amendment")
    root.set("session", amdt["congress"])
    root.set("chamber", amdt["amendment_type"][0])
    root.set("number", str(amdt["number"]))
    root.set("updated", utils.format_datetime(amdt["updated_at"]))

    make_node = utils.make_node

    if amdt.get("amends_bill", None):
        make_node(
            root,
            "amends",
            None,
            type=govtrack_type_codes[amdt["amends_bill"]["bill_type"]],
            number=str(amdt["amends_bill"]["number"]),
            sequence=str(amdt["house_number"])
            if amdt.get("house_number", None)
            else "",
        )
    elif amdt.get("amends_treaty", None):
        make_node(
            root,
            "amends",
            None,
            type="treaty",
            number=str(amdt["amends_treaty"]["number"]),
        )

    make_node(root, "status", amdt["status"], datetime=amdt["status_at"])

    if amdt["sponsor"] and amdt["sponsor"]["type"] == "person":
        v = amdt["sponsor"]["bioguide_id"]
        if not options.get("govtrack", False):
            make_node(root, "sponsor", None, bioguide_id=v)
        else:
            v = str(utils.translate_legislator_id("bioguide", v, "govtrack"))
            make_node(root, "sponsor", None, id=v)
    elif amdt["sponsor"] and amdt["sponsor"]["type"] == "committee":
        make_node(root, "sponsor", None, committee=amdt["sponsor"]["name"])
    else:
        make_node(root, "sponsor", None)

    make_node(root, "offered", None, datetime=amdt["introduced_at"])

    make_node(root, "description", amdt["description"] if amdt.get("description") else amdt["purpose"])
    if amdt.get("description"):
        make_node(root, "purpose", amdt["purpose"])

    actions = make_node(root, "actions", None)
    for action in amdt["actions"]:
        a = make_node(
            actions,
            action["type"] if action["type"] in ("vote",) else "action",
            None,
            datetime=action["acted_at"],
        )
        if action["type"] == "vote":
            a.set("how", action["how"])
            a.set("result", action["result"])
            if action.get("roll") != None:
                a.set("roll", str(action["roll"]))
        if action.get("text"):
            make_node(a, "text", action["text"])
        if action.get("in_committee"):
            make_node(a, "committee", None, name=action["in_committee"])
        for cr in action["references"]:
            make_node(a, "reference", None, ref=cr["reference"], label=cr["type"])

    return etree.tostring(root, pretty_print=True)


def build_amendment_id(amdt_type, amdt_number, congress):
    return "%s%s-%s" % (amdt_type, amdt_number, congress)


def amends_bill_for(amends_bill):
    from congress.tasks.bills import build_bill_id

    bill_id = build_bill_id(
        amends_bill["type"].lower(), amends_bill["number"], amends_bill["congress"]
    )
    return {
        "bill_id": bill_id,
        "bill_type": amends_bill["type"].lower(),
        "congress": int(amends_bill["congress"]),
        "number": int(amends_bill["number"]),
    }


def amends_amendment_for(amends_amdt):
    if amends_amdt is None:
        return None

    amdt_id = build_amendment_id(
        amends_amdt["type"].lower(), amends_amdt["number"], amends_amdt["congress"]
    )
    return {
        "amendment_id": amdt_id,
        "amendment_type": amends_amdt.get("type", "").lower(),
        "congress": int(amends_amdt.get("congress", "")),
        "number": int(amends_amdt.get("number", "")),
        "purpose": amends_amdt.get("purpose", ""),
        "description": amends_amdt.get("description", ""),
    }


def actions_for(action_list):
    if action_list is None:
        return []  # no actions
    actions = [action_for(action) for action in action_list["item"]]
    parse_amendment_actions(actions)
    return actions


def parse_amendment_actions(actions):
    for action in actions:
        # House Vote
        m = re.match(
            r"On agreeing to the .* amendments? (\(.*\) )?(?:as (?:modified|amended) )?(Agreed to|Failed) (without objection|by [^\.:]+|by (?:recorded vote|the Yeas and Nays): (\d+) - (\d+)(, \d+ Present)? \(Roll [nN]o. (\d+)\))\.",
            action["text"],
        )
        if m:
            action["where"] = "h"
            action["type"] = "vote"
            action["vote_type"] = "vote"

            if m.group(2) == "Agreed to":
                action["result"] = "pass"
            else:
                action["result"] = "fail"

            action["how"] = m.group(3)
            if "recorded vote" in m.group(3) or "the Yeas and Nays" in m.group(3):
                action["how"] = "roll"
                action["roll"] = int(m.group(7))

        # Senate Vote
        m = re.match(
            r"(Motion to table )?Amendment SA \d+(?:, .*?)? (as modified )?(agreed to|not agreed to) in Senate by ([^\.:\-]+|Yea-Nay( Vote)?. (\d+) - (\d+)(, \d+ Present)?. Record Vote Number: (\d+))\.",
            action["text"],
        )
        if m:
            action["type"] = "vote"
            action["vote_type"] = "vote"
            action["where"] = "s"

            if m.group(3) == "agreed to":
                action["result"] = "pass"
                if m.group(
                    1
                ):  # is a motion to table, so result is sort of reversed.... eeek
                    action["result"] = "fail"
            else:
                if m.group(
                    1
                ):  # is a failed motion to table, so this doesn't count as a vote on agreeing to the amendment
                    continue
                action["result"] = "fail"

            action["how"] = m.group(4)
            if "Yea-Nay" in m.group(4):
                action["how"] = "roll"
                action["roll"] = int(m.group(9))

        # Withdrawn
        m = re.match(r"Proposed amendment SA \d+ withdrawn in Senate", action["text"])
        if m:
            action["type"] = "withdrawn"


def amendment_status_for(amdt):
    status = "offered"
    status_date = amdt["introduced_at"]

    for action in amdt["actions"]:
        if action["type"] == "vote":
            status = action["result"]  # 'pass', 'fail'
            status_date = action["acted_at"]
        if action["type"] == "withdrawn":
            status = "withdrawn"
            status_date = action["acted_at"]

    return status, status_date


def sponsor_for(sponsor, amendment_type):
    if sponsor.get("bioguideId") is None:
        # A committee can sponsor an amendment!
        # Change e.g. "Rules Committee" to "House Rules" for the committee name,
        # for backwards compatibility.
        name = re.sub(
            r"(.*) Committee$",
            ("House" if (amendment_type[0] == "h") else "Senate") + r" \1",
            sponsor["name"],
        )
        return {
            "type": "committee",
            "name": name,
            # "committee_id": None, # TODO
        }

    return sponsor_for_bill(sponsor)


def output_for_amdt(amendment_id, format):
    amendment_type, number, congress = utils.split_bill_id(amendment_id)
    return "%s/%s/amendments/%s/%s%s/%s" % (
        utils.data_dir(),
        congress,
        amendment_type,
        amendment_type,
        number,
        "data.%s" % format,
    )



# -*- coding: utf-8 -*-
"""
A module that monkey-patches the output_bill method to push the bill identifier
onto a task queue after the data file has been written to disk. To use this
module, invoke the bills scraper with the --patch option like so:

  usc-run bills --patch=contrib.beanstalkd

You must include a 'beakstalk' section in config.yml with this structure
(though the values are up to you):

  beanstalk:
    connection:
      host: 'localhost'
      port: 11300
    tubes:
      bills: 'us_bills'
      amendments: 'us_amendments'
      votes: 'us_votes'
"""



import sys
import logging
import time
import traceback

from collections import Counter
from functools import wraps

import yaml
import beanstalkc

# The patch module is loaded after the task module is loaded, so all task
# modules are on the import path.
from congress.tasks import bills, amendment_info, vote_info


__all__ = [
    'patch',
    'process_bill_wrapper',
    'process_amendment_wrapper',
    'output_vote_wrapper'
]


_Connection = None
_Config = None


def init_guard(reconnect=False):
    global _Config, _Connection
    if _Config is None:
        with open('config.yml', 'r') as conffile:
            config = yaml.load(conffile, yaml.SafeLoader)
            assert 'beanstalk' in config
            assert 'connection' in config['beanstalk']
            assert 'host' in config['beanstalk']['connection']
            assert 'port' in config['beanstalk']['connection']
            assert 'tubes' in config['beanstalk']
            assert 'bills' in config['beanstalk']['tubes']
            assert 'amendments' in config['beanstalk']['tubes']
            assert 'votes' in config['beanstalk']['tubes']
            tube_names = list(config['beanstalk']['tubes'].values())
            assert max(Counter(tube_names).values()) == 1, 'Must use unique beanstalk tube names.'
            _Config = config['beanstalk']
    if _Connection is None or reconnect is True:
        conn = beanstalkc.Connection(**_Config['connection'])
        assert conn is not None
        _Connection = conn
    return (_Connection, _Config)


def process_bill_wrapper(process_bill):
    @wraps(process_bill)
    def _process_bill(bill, options, *args, **kwargs):
        orig_result = process_bill(bill, options, *args, **kwargs)

        (conn, config) = init_guard()
        for _ in range(2):
            try:
                conn.use(config['tubes']['bills'])
                conn.put(bill)
                logging.warn("Queued {} to beanstalkd.".format(bill))
                break
            except beanstalkc.SocketError:
                logging.warn("Lost connection to beanstalkd. Attempting to reconnect.")
                (conn, config) = init_guard(reconnect=True)
            except Exception as e:
                logging.warn("Ignored exception while queueing bill to beanstalkd: {0} {1}".format(str(type(e)), str(e)))
                traceback.print_exc()
                break

        return orig_result

    return _process_bill


def process_amendment_wrapper(process_amendment):
    @wraps(process_amendment)
    def _process_amendment(amdt_dict, bill_id, options, *args, **kwargs):
        orig_result = process_amendment(amdt_dict, bill_id, options, *args, **kwargs)
        amdt = amendment_info.build_amendment_id(amdt_dict['type'].lower(), amdt_dict['number'], amdt_dict['congress'])

        (conn, config) = init_guard()
        for _ in range(2):
            try:
                conn.use(config['tubes']['amendments'])
                conn.put(str(amdt))
                logging.warn("Queued {} to beanstalkd.".format(amdt))
                break
            except beanstalkc.SocketError:
                logging.warn("Lost connection to beanstalkd. Attempting to reconnect.")
                (conn, config) = init_guard(reconnect=True)
            except Exception as e:
                logging.warn("Ignored exception while queueing amendment to beanstalkd: {0} {1}".format(str(type(e)), str(e)))
                traceback.print_exc()
                break

        return orig_result

    return _process_amendment


def output_vote_wrapper(output_vote):
    @wraps(output_vote)
    def _output_vote(vote, options, *args, **kwargs):
        orig_result = output_vote(vote, options, *args, **kwargs)

        (conn, config) = init_guard()
        for _ in range(2):
            try:
                conn.use(config['tubes']['votes'])
                conn.put(vote['vote_id'])
                logging.warn('Queued {} to beanstalkd.'.format(vote['vote_id']))
                break
            except beanstalkc.SocketError:
                logging.warn('Lost connection to beanstalkd. Attempting to reconnect.')
                (conn, config) = init_guard(reconnect=True)
            except Exception as e:
                logging.warn('Ignored exception while queueing vote to beanstalkd: {0} {1}'.format(str(type(e)), str(e)))
                traceback.print_exc()
                break

        return orig_result

    return _output_vote


def patch(task_name):
    bills.process_bill = process_bill_wrapper(bills.process_bill)
    amendment_info.process_amendment = process_amendment_wrapper(amendment_info.process_amendment)
    vote_info.output_vote = output_vote_wrapper(vote_info.output_vote)


# Avoid scraping if the beanstalk config is invalid.
try:
    init_guard()
except AssertionError:
    print(__doc__)
    sys.exit(1)



from congress.tasks import utils
import logging
import re
import json
from lxml import etree
import time
import datetime
import os
import os.path


def fetch_vote(vote_id, options):
    logging.info("\n[%s] Fetching..." % vote_id)

    vote_chamber, vote_number, vote_congress, vote_session_year = utils.split_vote_id(
        vote_id
    )

    if vote_chamber == "h":
        url = "https://clerk.house.gov/evs/%s/roll%03d.xml" % (
            vote_session_year,
            int(vote_number),
        )
    else:
        session_num = (
            int(vote_session_year)
            - utils.get_congress_first_year(int(vote_congress))
            + 1
        )
        url = (
            "https://www.senate.gov/legislative/LIS/roll_call_votes/vote%d%d/vote_%d_%d_%05d.xml"
            % (
                int(vote_congress),
                session_num,
                int(vote_congress),
                session_num,
                int(vote_number),
            )
        )

    # fetch vote XML page
    body = utils.download(
        url,
        "%s/votes/%s/%s%s/%s%s.xml"
        % (
            vote_congress,
            vote_session_year,
            vote_chamber,
            vote_number,
            vote_chamber,
            vote_number,
        ),
        utils.merge(options, {"binary": True}),
    )

    if not body:
        return {"saved": False, "ok": False, "reason": "failed to download"}

    if options.get("download_only", False):
        return {"saved": False, "ok": True, "reason": "requested download only"}

    if b"This vote was vacated" in body:
        # Vacated votes: 2011-484, 2012-327, ...
        # Remove file, since it may previously have existed with data.
        for f in (output_for_vote(vote_id, "json"), output_for_vote(vote_id, "xml")):
            if os.path.exists(f):
                os.unlink(f)
        return {"saved": False, "ok": True, "reason": "vote was vacated"}

    if b"roll-call-vote-not-available.htm" in body:
        # Vote showed up in the xml list of votes but no roll call file has been
        # created yet, or script was given a completely invalid vote number.
        # In the case of very recent votes, this error should be temporary.
        return {"saved": False, "ok": True, "reason": "roll call vote not available"}

    if b"roll-call-vote-not-available.htm" in body:
        # Vote showed up in the xml list of votes but no roll call file has been
        # created yet, or script was given a completely invalid vote number.
        # In the case of very recent votes, this error should be temporary.
        return {"saved": False, "ok": True, "reason": "roll call vote not available"}

    dom = etree.fromstring(body)

    vote = {
        "vote_id": vote_id,
        "chamber": vote_chamber,
        "congress": int(vote_congress),
        "session": vote_session_year,
        "number": int(vote_number),
        "updated_at": datetime.datetime.fromtimestamp(time.time()),
        "source_url": url,
    }

    # do the heavy lifting

    if vote_chamber == "h":
        parse_house_vote(dom, vote)
    elif vote_chamber == "s":
        parse_senate_vote(dom, vote)

    # output and return

    output_vote(vote, options)

    return {"ok": True, "saved": True}


def output_vote(vote, options, id_type=None):
    logging.info("[%s] Writing to disk..." % vote["vote_id"])

    # output JSON - so easy!
    data = json.dumps(vote, sort_keys=True, indent=2, default=utils.format_datetime)
    destination = output_for_vote(vote["vote_id"], "json")
    utils.write(data, output_for_vote(vote["vote_id"], "json"), options=options)
    cb = options.get("callback", None)
    if cb and callable(cb):
        cb(destination)

    # What kind of IDs are we passed for Members of Congress?
    # For current data, we infer from the chamber. For historical data from voteview,
    # we're passed the type in id_type, which is set to "bioguide".
    if not id_type:
        id_type = "bioguide" if vote["chamber"] == "h" else "lis"

    # output XML
    root = etree.Element("roll")

    root.set("where", "house" if vote["chamber"] == "h" else "senate")
    root.set("session", str(vote["congress"]))
    root.set("year", str(vote["date"].year))
    root.set("roll", str(vote["number"]))
    if "voteview" in vote["source_url"]:
        root.set("source", "keithpoole")
    else:
        root.set("source", "house.gov" if vote["chamber"] == "h" else "senate.gov")

    root.set("datetime", utils.format_datetime(vote["date"]))
    root.set("updated", utils.format_datetime(vote["updated_at"]))

    def get_votes(option):
        return len(vote["votes"].get(option, []))

    root.set("aye", str(get_votes("Yea") + get_votes("Aye")))
    root.set("nay", str(get_votes("Nay") + get_votes("No")))
    root.set("nv", str(get_votes("Not Voting")))
    root.set("present", str(get_votes("Present")))

    utils.make_node(root, "category", vote["category"])
    utils.make_node(root, "type", vote["type"])
    utils.make_node(root, "question", vote["question"])
    utils.make_node(root, "required", vote["requires"])
    utils.make_node(root, "result", vote["result"])

    if vote.get("bill"):
        govtrack_type_codes = {
            "hr": "h",
            "s": "s",
            "hres": "hr",
            "sres": "sr",
            "hjres": "hj",
            "sjres": "sj",
            "hconres": "hc",
            "sconres": "sc",
        }
        utils.make_node(
            root,
            "bill",
            None,
            session=str(vote["bill"]["congress"]),
            type=govtrack_type_codes[vote["bill"]["type"]],
            number=str(vote["bill"]["number"]),
        )

    if "amendment" in vote:
        n = utils.make_node(root, "amendment", None)
        if vote["amendment"]["type"] == "s":
            n.set("ref", "regular")
            n.set("session", str(vote["congress"]))
            n.set("number", "s" + str(vote["amendment"]["number"]))
        elif vote["amendment"]["type"] == "h-bill":
            n.set("ref", "bill-serial")
            n.set("session", str(vote["congress"]))
            n.set("number", str(vote["amendment"]["number"]))

    # well-known keys for certain vote types: +/-/P/0
    option_keys = {
        "Aye": "+",
        "Yea": "+",
        "Nay": "-",
        "No": "-",
        "Present": "P",
        "Not Voting": "0",
        "Guilty": "+",
        "Not Guilty": "-",
    }

    # preferred order of output: ayes, nays, present, then not voting, and similarly for guilty/not-guilty
    # and handling other options like people's names for votes for the Speaker.
    option_sort_order = (
        "Aye",
        "Yea",
        "Guilty",
        "No",
        "Nay",
        "Not Guilty",
        "OTHER",
        "Present",
        "Not Voting",
    )
    options_list = sorted(
        vote["votes"].keys(),
        key=lambda o: (
            option_sort_order.index(o)
            if o in option_sort_order
            else option_sort_order.index("OTHER")
        ),
    )
    for option in options_list:
        if option not in option_keys:
            option_keys[option] = option
        utils.make_node(root, "option", option, key=option_keys[option])

    for option in options_list:
        for v in vote["votes"][option]:
            # Rep-elect Letlow is included as not voting in the first House vote of the 117th Congress
            # where the clerk calls a quorum roll call. But because Letlow had died prior to this date,
            # he is not represented in congress-legislators and has no GovTrack-id, and so we cannot
            # represent this record in the data.
            if (
                isinstance(v, dict)
                and v["id"] == "L000555"
                and options.get("govtrack", False)
            ):
                continue

            n = utils.make_node(root, "voter", None)
            if v == "VP":
                n.set("id", "0")
                n.set("VP", "1")
            elif not options.get("govtrack", False):
                n.set("id", str(v["id"]))
            else:
                n.set(
                    "id",
                    str(utils.translate_legislator_id(id_type, v["id"], "govtrack")),
                )
            n.set("vote", option_keys[option])
            n.set("value", option)
            if v != "VP":
                n.set("state", v["state"])
                if v.get("voteview_votecode_extra") is not None:
                    n.set("voteview_votecode_extra", v["voteview_votecode_extra"])

    xmloutput = etree.tostring(root, pretty_print=True, encoding="unicode")

    # mimick two hard line breaks in GovTrack's legacy output to ease running diffs
    xmloutput = re.sub('(source=".*?") ', r"\1\n  ", xmloutput)
    xmloutput = re.sub('(updated=".*?") ', r"\1\n  ", xmloutput)

    utils.write(xmloutput, output_for_vote(vote["vote_id"], "xml"), options=options)


def output_for_vote(vote_id, format):
    vote_chamber, vote_number, vote_congress, vote_session_year = utils.split_vote_id(
        vote_id
    )
    return "%s/%s/votes/%s/%s%s/%s" % (
        utils.data_dir(),
        vote_congress,
        vote_session_year,
        vote_chamber,
        vote_number,
        "data.%s" % format,
    )


def parse_senate_vote(dom, vote):
    def parse_date(d):
        return datetime.datetime.strptime(d, "%B %d, %Y, %I:%M %p")

    vote["date"] = parse_date(dom.xpath("string(vote_date)"))
    if len(dom.xpath("modify_date")) > 0:
        vote["record_modified"] = parse_date(
            dom.xpath("string(modify_date)")
        )  # some votes like s1-110.2008 don't have a modify_date
    vote["question"] = str(dom.xpath("string(vote_question_text)"))
    if vote["question"] == "":
        vote["question"] = str(dom.xpath("string(question)"))  # historical votes?
    vote["type"] = str(dom.xpath("string(vote_question)"))
    if vote["type"] == "":
        vote["type"] = vote["question"]
    vote["type"] = normalize_vote_type(vote["type"])
    vote["category"] = get_vote_category(vote["type"])
    vote["subject"] = str(dom.xpath("string(vote_title)"))
    vote["requires"] = str(dom.xpath("string(majority_requirement)"))
    vote["result_text"] = str(dom.xpath("string(vote_result_text)"))
    vote["result"] = str(dom.xpath("string(vote_result)"))

    # Senate cloture votes have consistently bad vote_question_text values: They don't say what the cloture
    # was about specifically, just what bill was relevant. So cloture on an amendment just appears as
    # cloture on the bill. The vote_title text is correctly specific in those cases. So swap the two fields
    # in those cases so that our 'question' field is reliably a good title, and subject provides additional
    # information. Check that the subject is non-empty before using it, just in case. Example:
    #  "question": "On the Cloture Motion H.R. 2578"
    #  "subject": "Motion to Invoke Cloture on the Motion to Commit H.R. 2578 with instructions (Amdt. No. 4750)"
    # https://www.senate.gov/legislative/LIS/roll_call_votes/vote1142/vote_114_2_00104.xml
    if "Cloture" in vote["question"] and vote["subject"]:
        x = vote["question"]
        vote["question"] = vote["subject"]
        vote["subject"] = x

    # "Motion to Proceed to Legislative Session" is also consistently buried and weirdly attached to a nomination.
    # Swap the fields in that case too and unlink it from the nomination because it's confusing.
    # (Should we do the swap for all motions to proceed?)
    # https://www.senate.gov/legislative/LIS/roll_call_votes/vote1151/vote_115_1_00049.xml
    elif "Legislative Session" in vote["subject"]:
        x = vote["question"]
        vote["question"] = vote["subject"]
        vote["subject"] = x
        for n in dom.xpath("document/document_type"):
            n.text = None

    bill_types = {
        "S.": "s",
        "S.Con.Res.": "sconres",
        "S.J.Res.": "sjres",
        "S.Res.": "sres",
        "H.R.": "hr",
        "H.Con.Res.": "hconres",
        "H.J.Res.": "hjres",
        "H.Res.": "hres",
    }

    if str(dom.xpath("string(document/document_type)")):
        if dom.xpath("string(document/document_type)") == "PN":
            vote["nomination"] = {
                "number": str(dom.xpath("string(document/document_number)")),
                "title": str(dom.xpath("string(document/document_title)")),
            }
            vote["question"] += ": " + vote["nomination"]["title"]
        elif dom.xpath("string(document/document_type)") == "Treaty Doc.":
            vote["treaty"] = {
                "title": str(dom.xpath("string(document/document_title)")),
            }
        elif str(dom.xpath("string(document/document_type)")) in bill_types:
            vote["bill"] = {
                "congress": int(
                    dom.xpath("number(document/document_congress|congress)")
                ),  # some historical files don't have document/document_congress so take the first of document/document_congress or the top-level congress element as a fall-back
                "type": bill_types[str(dom.xpath("string(document/document_type)"))],
                "number": int(dom.xpath("number(document/document_number)")),
                "title": str(dom.xpath("string(document/document_title)")),
            }
        else:
            # s294-115.2017 through s302-115.2017 have S.Amdt. in document_type,
            # but it probably should be empty since <amendment> is filled in and
            # the rest of <document> is blank.
            pass

    if str(dom.xpath("string(amendment/amendment_number)")):
        m = re.match(
            r"^S.Amdt. (\d+)", str(dom.xpath("string(amendment/amendment_number)"))
        )
        if m:
            vote["amendment"] = {
                "type": "s",
                "number": int(m.group(1)),
                "purpose": str(dom.xpath("string(amendment/amendment_purpose)")),
            }

        amendment_to = str(dom.xpath("string(amendment/amendment_to_document_number)"))
        if "Treaty" in amendment_to:
            treaty, number = amendment_to.split("-")
            vote["treaty"] = {
                "congress": vote["congress"],
                "number": number,
            }
        elif " " in amendment_to:
            bill_type, bill_number = amendment_to.split(" ")
            vote["bill"] = {
                "congress": vote["congress"],
                "type": bill_types[bill_type],
                "number": int(bill_number),
                "title": str(
                    dom.xpath("string(amendment/amendment_to_document_short_title)")
                ),
            }
        else:
            # Senate votes:
            # 102nd Congress, 2nd session (1992): 247, 248, 250; 105th Congress, 2nd session (1998): 106 through 116; 108th Congress, 1st session (2003): 41, 42
            logging.warn(
                "Amendment without corresponding bill info in %s " % vote["vote_id"]
            )

    # Count up the votes.
    vote["votes"] = {}

    def add_vote(vote_option, voter):
        if vote_option == "Present, Giving Live Pair":
            vote_option = "Present"
        vote["votes"].setdefault(vote_option, []).append(voter)

        # In the 101st Congress, 1st session (1989), votes 133 through 136 lack lis_member_id nodes.
        if voter != "VP" and voter["id"] == "":
            voter["id"] = utils.lookup_legislator(
                vote["congress"],
                "sen",
                voter["last_name"],
                voter["state"],
                voter["party"],
                vote["date"],
                "lis",
            )
            if voter["id"] == None:
                logging.error(
                    "[%s] Missing lis_member_id and name lookup failed for %s"
                    % (vote["vote_id"], voter["last_name"])
                )
                raise Exception(
                    "Could not find ID for %s (%s-%s)"
                    % (voter["last_name"], voter["state"], voter["party"])
                )
            else:
                logging.info(
                    "[%s] Missing lis_member_id, falling back to name lookup for %s"
                    % (vote["vote_id"], voter["last_name"])
                )

    # Ensure the options are noted, even if no one votes that way.
    if str(dom.xpath("string(question)")) == "Guilty or Not Guilty":
        vote["votes"]["Guilty"] = []
        vote["votes"]["Not Guilty"] = []
    else:
        vote["votes"]["Yea"] = []
        vote["votes"]["Nay"] = []
    vote["votes"]["Present"] = []
    vote["votes"]["Not Voting"] = []

    # VP tie-breaker?
    if str(dom.xpath("string(tie_breaker/by_whom)")):
        add_vote(str(dom.xpath("string(tie_breaker/tie_breaker_vote)")), "VP")

    for member in dom.xpath("members/member"):
        add_vote(
            str(member.xpath("string(vote_cast)")),
            {
                "id": str(member.xpath("string(lis_member_id)")),
                "state": str(member.xpath("string(state)")),
                "party": str(member.xpath("string(party)")),
                "display_name": str(member.xpath("string(member_full)")),
                "first_name": str(member.xpath("string(first_name)")),
                "last_name": str(member.xpath("string(last_name)")),
            },
        )


def parse_house_vote(dom, vote):
    def parse_date(d):
        d = d.strip()
        if " " in d:
            return datetime.datetime.strptime(d, "%d-%b-%Y %I:%M %p")
        else:  # some votes have no times?
            print(vote)
            return datetime.datetime.strptime(d, "%d-%b-%Y")

    vote["date"] = parse_date(
        str(dom.xpath("string(vote-metadata/action-date)"))
        + " "
        + str(dom.xpath("string(vote-metadata/action-time)"))
    )
    vote["question"] = str(dom.xpath("string(vote-metadata/vote-question)"))
    vote["type"] = str(dom.xpath("string(vote-metadata/vote-question)"))
    vote["type"] = normalize_vote_type(vote["type"])
    if str(dom.xpath("string(vote-metadata/vote-desc)")).startswith("Impeaching "):
        vote["category"] = "impeachment"
    else:
        vote["category"] = get_vote_category(vote["question"])
    vote["subject"] = str(dom.xpath("string(vote-metadata/vote-desc)"))
    if not vote["subject"]:
        del vote["subject"]

    vote_types = {
        "YEA-AND-NAY": "1/2",
        "2/3 YEA-AND-NAY": "2/3",
        "3/5 YEA-AND-NAY": "3/5",
        "1/2": "1/2",
        "2/3": "2/3",
        "QUORUM": "QUORUM",
        "RECORDED VOTE": "1/2",
        "2/3 RECORDED VOTE": "2/3",
        "3/5 RECORDED VOTE": "3/5",
    }
    vote["requires"] = vote_types.get(
        str(dom.xpath("string(vote-metadata/vote-type)")), "unknown"
    )

    vote["result_text"] = str(dom.xpath("string(vote-metadata/vote-result)"))
    vote["result"] = str(dom.xpath("string(vote-metadata/vote-result)"))

    bill_num = str(dom.xpath("string(vote-metadata/legis-num)"))
    if bill_num not in ("", "QUORUM", "JOURNAL", "MOTION", "ADJOURN") and not re.match(
        r"QUORUM \d+$", bill_num
    ):
        bill_types = {
            "S": "s",
            "S CON RES": "sconres",
            "S J RES": "sjres",
            "S RES": "sres",
            "H R": "hr",
            "H CON RES": "hconres",
            "H J RES": "hjres",
            "H RES": "hres",
        }
        try:
            bill_type, bill_number = bill_num.rsplit(" ", 1)
            vote["bill"] = {
                "congress": vote["congress"],
                "type": bill_types[bill_type],
                "number": int(bill_number),
            }
        except (
            ValueError
        ):  # rsplit failed, i.e. there is no space in the legis-num field
            raise Exception("Unhandled bill number in the legis-num field")

    if str(dom.xpath("string(vote-metadata/amendment-num)")):
        vote["amendment"] = {
            "type": "h-bill",
            "number": int(str(dom.xpath("string(vote-metadata/amendment-num)"))),
            "author": str(dom.xpath("string(vote-metadata/amendment-author)")),
        }

    # Assemble a complete question from the vote type, amendment, and bill number.
    if "amendment" in vote and "bill" in vote:
        vote["question"] += ": Amendment %s to %s" % (
            vote["amendment"]["number"],
            str(dom.xpath("string(vote-metadata/legis-num)")),
        )
    elif "amendment" in vote:
        vote["question"] += (
            ": Amendment %s to [unknown bill]" % vote["amendment"]["number"]
        )
    elif "bill" in vote:
        vote["question"] += ": " + str(dom.xpath("string(vote-metadata/legis-num)"))
        if "subject" in vote:
            vote["question"] += " " + vote["subject"]
    elif "subject" in vote:
        vote["question"] += ": " + vote["subject"]

    # Count up the votes.
    vote["votes"] = {}  # by vote type

    def add_vote(vote_option, voter):
        vote["votes"].setdefault(vote_option, []).append(voter)

    # Ensure the options are noted, even if no one votes that way.
    if (
        str(dom.xpath("string(vote-metadata/vote-question)"))
        == "Election of the Speaker"
    ):
        for n in dom.xpath("vote-metadata/vote-totals/totals-by-candidate/candidate"):
            vote["votes"][n.text] = []
    elif str(dom.xpath("string(vote-metadata/vote-question)")) == "Call of the House":
        for n in dom.xpath("vote-metadata/vote-totals/totals-by-candidate/candidate"):
            vote["votes"][n.text] = []
    elif "YEA-AND-NAY" in dom.xpath("string(vote-metadata/vote-type)"):
        vote["votes"]["Yea"] = []
        vote["votes"]["Nay"] = []
        vote["votes"]["Present"] = []
        vote["votes"]["Not Voting"] = []
    else:
        vote["votes"]["Aye"] = []
        vote["votes"]["No"] = []
        vote["votes"]["Present"] = []
        vote["votes"]["Not Voting"] = []

    for member in dom.xpath("vote-data/recorded-vote"):
        display_name = str(member.xpath("string(legislator)"))
        state = str(member.xpath("string(legislator/@state)"))
        party = str(member.xpath("string(legislator/@party)"))
        vote_cast = str(member.xpath("string(vote)"))
        bioguideid = str(member.xpath("string(legislator/@name-id)"))
        add_vote(
            vote_cast,
            {
                "id": bioguideid,
                "state": state,
                "party": party,
                "display_name": display_name,
            },
        )

    # Through the 107th Congress and sporadically in more recent data, the bioguide field
    # is not present. Look up the Members' bioguide IDs by name/state/party/date. This works
    # reasonably well, but there are many gaps. When there's a gap, it raises an exception
    # and the file is not saved.
    #
    # Take into account that the vote may list both a "Smith" and a "Smith, John". Resolve
    # "Smith" by process of elimination, i.e. he must not be whoever "Smith, John" resolved
    # to. To do that, process the voters from longest specified display name to shortest.
    #
    # One example of a sporadic case is 108th Congress, 2nd session (2004), votes 405 through
    # 544, where G.K. Butterfield's bioguide ID is 000000. It should have been B001251.
    # See https://github.com/unitedstates/congress/issues/46.

    seen_ids = set()
    all_voters = sum(vote["votes"].values(), [])
    all_voters.sort(
        key=lambda v: len(v["display_name"]), reverse=True
    )  # process longer names first
    for v in all_voters:
        if v["id"] not in ("", "0000000"):
            continue

        # here are wierd cases from h610-103.1993 that confound our name lookup since it has the wrong state abbr
        if v["state"] == "XX":
            for st in ("PR", "AS", "GU", "VI", "DC"):
                if v["display_name"].endswith(" (%s)" % st):
                    v["state"] = st

        # get the last name without the state abbreviation in parenthesis, if it is present
        display_name = v["display_name"].strip()
        ss = " (%s)" % v["state"]
        if display_name.endswith(ss):
            display_name = display_name[: -len(ss)].strip()

        # wrong party in upstream data
        if vote["vote_id"] == "h2-106.1999" and display_name == "Hastert":
            v["id"] = "H000323"
            continue

        # dead man recorded as Not Voting (he died the day before, so none of our roles match the vote date)
        if (
            vote["vote_id"] == "h306-106.1999"
            and display_name == "Brown"
            and v["state"] == "CA"
        ):
            v["id"] = "B000918"
            continue

        # look up ID
        v["id"] = utils.lookup_legislator(
            vote["congress"],
            "rep",
            display_name,
            v["state"],
            v["party"],
            vote["date"],
            "bioguide",
            exclude=seen_ids,
        )

        if v["id"] == None:
            logging.error(
                "[%s] Missing bioguide ID and name lookup failed for %s (%s-%s on %s)"
                % (vote["vote_id"], display_name, v["state"], v["party"], vote["date"])
            )
            raise Exception(
                "No bioguide ID for %s (%s-%s)" % (display_name, v["state"], v["party"])
            )
        else:
            if vote["congress"] > 107:
                logging.warn(
                    "[%s] Used name lookup for %s because bioguide ID was missing."
                    % (vote["vote_id"], v["display_name"])
                )
            seen_ids.add(v["id"])


def normalize_vote_type(vote_type):
    # Takes the "type" field of a House or Senate vote and returns a normalized
    # version of the same, as best as possible.

    # note that these allow .* after each pattern, so some things look like
    # no-ops but they are really truncating the type after the specified text.
    mapping = (
        (
            r"^On the Resolution of Ratification.*",
            "On the Resolution of Ratification",
        ),  # order matters so must go before other resolutions
        (
            r"On (Agreeing to )?the (Joint |Concurrent )?Resolution",
            "On the $2Resolution",
        ),
        (r"On (Agreeing to )?the Conference Report", "On the Conference Report"),
        (r"On (Agreeing to )?the (En Bloc )?Amendments?", "On the Amendment"),
        (r"On (?:the )?Motion to Recommit", "On the Motion to Recommit"),
        (
            r"(On Motion to )?(Concur in|Concurring|On Concurring|Agree to|On Agreeing to) (the )?Senate (Amendment|amdt|Adt)s?",
            "Concurring in the Senate Amendment",
        ),
        (
            r"(On Motion to )?Suspend (the )?Rules and (Agree|Concur|Pass)(, As Amended)",
            "On Motion to Suspend the Rules and $3$4",
        ),
        (
            r"Will the House Now Consider the Resolution|On (Question of )?Consideration of the Resolution",
            "On Consideration of the Resolution",
        ),
        (r"On (the )?Motion to Adjourn", "On the Motion to Adjourn"),
        (r"On (the )?Cloture Motion", "On the Cloture Motion"),
        (r"On Cloture on the Motion to Proceed", "On the Cloture Motion"),
        (r"On (the )?Nomination", "On the Nomination"),
        (r"On Passage( of the Bill|$)", "On Passage of the Bill"),
        (r"On (the )?Motion to Proceed", "On the Motion to Proceed"),
        (
            r"On the Motion \(Motion to ((Recede )from the Senate Amendment to \S+ \d+ (and ))?Concur( with Further Amendment)?",
            "On the Motion to $2$3Concur$4",
        ),
        (r"On the Motion \(Motion to (.*)\)$", "On the Motion to $1"),
    )

    for regex, replacement in mapping:
        m = re.match(regex, vote_type, re.I)
        if m:
            if m.groups():
                for i, val in enumerate(m.groups()):
                    replacement = replacement.replace(
                        "$%d" % (i + 1), val if val else ""
                    )
            return replacement

    return vote_type


def get_vote_category(vote_question):
    # Takes the type/question field of a House or Senate vote and returns a normalized
    # category for the vote type.
    #
    # Based on Eric's vote_type_for function in sunlightlabs/congress.

    mapping = (
        # empty text (historical data)
        (r"^$", "unknown"),
        # common
        (r"^On Overriding the Veto", "veto-override"),
        (r"^On Presidential Veto", "veto-override"),
        (
            r"Objections of the President (To The Contrary )?Not ?Withstanding",
            "veto-override",
        ),  # order matters so must go before bill passage
        (r"^On Passage", "passage"),
        (
            r"^On the Resolution of Ratification.*",
            "treaty",
        ),  # order matters so must go before other resolutions
        (r"^On (Agreeing to )?the (Joint |Concurrent )?Resolution", "passage"),
        (r"^On (Agreeing to )?the Conference Report", "passage"),
        (r"^On (Agreeing to )?the (En Bloc )?Amendments?", "amendment"),
        # senate only
        (r"cloture", "cloture"),
        (r"^On the Nomination", "nomination"),
        (
            r"^Guilty or Not Guilty",
            "conviction",
        ),  # was "impeachment" in sunlightlabs/congress but that's not quite right
        (r"^On (?:the )?Motion to Recommit", "recommit"),
        (r"^On the Motion to Concur", "passage"),
        (
            r"^On the Motion to Recede and Concur with Further Amendment",
            "passage",
        ),  # this is a normalized type that is returned by a function above
        # house only
        (
            r"^(On Motion (to|that the House) )?(Concur in|Concurring|Concurring in|On Concurring|On Concurring in|Agree to|On Agreeing to) (the )?Senate (Amendment|amdt|Adt)s?",
            "passage",
        ),
        (
            r"^(On Motion to )?Suspend (the )?Rules and (Agree|Concur|Pass)",
            "passage-suspension",
        ),
        (r"^Call of the House$", "quorum"),
        (r"^Call by States$", "quorum"),
        (r"^Election of the Speaker$", "leadership"),
        # various procedural things
        # order matters, so these must go last
        (r"^On Ordering the Previous Question", "procedural"),
        (r"^On Approving the Journal", "procedural"),
        (
            r"^Will the House Now Consider the Resolution|On (Question of )?Consideration of the Resolution",
            "procedural",
        ),
        (r"^On (the )?Motion to Adjourn", "procedural"),
        (r"Authoriz(e|ing) Conferees", "procedural"),
        (r"On the Point of Order|Sustaining the Ruling of the Chair", "procedural"),
        (r"^On .*Motion ", "procedural"),  # $1 is a name like "Broun of Georgia"
        (r"^On the Decision of the Chair", "procedural"),
        (r"^Whether the Amendment is Germane", "procedural"),
        (r"^Table Appeal of the Ruling of the Chair", "procedural"),
    )

    for regex, category in mapping:
        if re.search(regex, vote_question, re.I):
            return category

    # unhandled
    logging.warn("Unhandled vote question: %s" % vote_question)
    return "unknown"



#!/usr/bin/env python

import sys
import os
import traceback
import pprint as pp
import logging
import importlib

# set global HTTP timeouts to 10 seconds
import socket

def main():
    socket.setdefaulttimeout(10)

    CONGRESS_ROOT = os.path.dirname(os.path.abspath(__file__))

    # name of the task comes first
    task_name = sys.argv[1]

    # parse any command line flags off
    options = {}
    for arg in sys.argv[2:]:
        if arg.startswith("--"):

            if "=" in arg:
                key, value = arg.split('=')
            else:
                key, value = arg, True

            key = key.split("--")[1]
            if value == 'True':
                value = True
            elif value == 'False':
                value = False
            options[key.lower()] = value


    # configure logging
    if options.get('debug', False):
        log_level = "debug"
    else:
        log_level = options.get("log", "warn")

    if log_level not in ["debug", "info", "warn", "error"]:
        print("Invalid log level (specify: debug, info, warn, error).")
        sys.exit(1)

    if options.get('timestamps', False):
        logging.basicConfig(format='%(asctime)s %(message)s', level=log_level.upper())
    else:
        logging.basicConfig(format='%(message)s', level=log_level.upper())


    sys.path.append(os.path.join(CONGRESS_ROOT, "tasks"))
    import utils

    try:
        task_mod = __import__(task_name)

        if 'patch' in options:
            patch_mod = importlib.import_module(options['patch'])
            patch_func = getattr(patch_mod, 'patch', None)
            if patch_func is None:
                logging.error("You specified a --patch argument but the {} module does not contain a 'patch' function.".format(options['patch']))
                sys.exit(1)
            elif not callable(patch_func):
                logging.error("You specified a --patch argument but {}.patch is not callable".format(options['patch']))
                sys.exit(1)
            else:
                patch_mod.patch(task_name)

        task_mod.run(options)
    except Exception as exception:
        utils.admin(exception)

if __name__ == "__main__":
    main()



from congress.tasks import utils
import json
from iso8601 import iso8601
import datetime
import os
import os.path
import re
import urllib.parse
import time
import datetime
from lxml import html, etree
import logging

from congress.tasks import vote_info


def run(options):
    # Parse the options and fill to_fetch with vote ids.

    vote_id = options.get("vote_id", None)
    if vote_id:
        # Just fetch the specified vote.
        vote_chamber, vote_number, congress, session_year = utils.split_vote_id(vote_id)
        sessions = [(congress, session_year)]
        to_fetch = [vote_id]
    else:
        if options.get("congress", None):
            if options.get("session", None):
                # Fetch for one session given with the congress and session options.
                sessions = [(options.get("congress"), options.get("session"))]
            else:
                # Fetch for both sessions of a Congress except sessions that are in the future.
                sessions = []
                first_session_year = utils.get_congress_first_year(
                    options.get("congress")
                )
                for y in (first_session_year, first_session_year + 1):
                    if y > datetime.datetime.now().year:
                        continue  # this session hasn't started yet
                    sessions.append((options.get("congress"), str(y)))
        elif options.get("sessions", None):
            # Fetch for multiple sessions, e.g. sessions=117.2021,117.2022
            sessions = [
                session.split(".") for session in options.get("sessions").split(",")
            ]
        else:
            # Fetch for the current session.
            sessions = [
                (
                    utils.current_congress(),
                    options.get("session", str(utils.current_legislative_year())),
                )
            ]

        # Get the list of votes.
        to_fetch = []
        for congress, session_year in sessions:
            if options.get("chamber", None) in ("house", None):
                to_fetch += vote_ids_for_house(congress, session_year, options) or []
            if options.get("chamber", None) in ("senate", None):
                to_fetch += vote_ids_for_senate(congress, session_year, options) or []

        if not to_fetch:
            if not options.get("fast", False):
                logging.error("Error figuring out which votes to download, aborting.")
            else:
                logging.warn("No new or recent votes.")
            return None

        limit = options.get("limit", None)
        if limit:
            to_fetch = to_fetch[: int(limit)]

    if options.get("pages_only", False):
        return None

    logging.warn(
        "Going to fetch %i votes from congress/session %s"
        % (len(to_fetch), ", ".join(str(cs) for cs in sessions))
    )

    utils.process_set(to_fetch, vote_info.fetch_vote, options)


# page through listing of House votes of a particular congress and session


def vote_ids_for_house(congress, session_year, options):
    vote_ids = []

    index_page = "https://clerk.house.gov/evs/%s/index.asp" % session_year
    group_page = r"ROLL_(\d+)\.asp"
    link_pattern = (
        r"http://clerk.house.gov/cgi-bin/vote.asp\?year=%s&rollnumber=(\d+)"
        % session_year
    )

    # download index page, find the matching links to the paged listing of votes
    download_options = options.copy()
    if options.get("fast", False):
        download_options["force"] = True  # Force fresh download when using --fast

    page = utils.download(
        index_page,
        "%s/votes/%s/pages/house.html" % (congress, session_year),
        download_options,
    )

    if not page:
        logging.error("Couldn't download House vote index page, skipping")
        return None

    # extract matching links
    doc = html.document_fromstring(page)
    links = doc.xpath(
        "//a[re:match(@href, '%s')]" % group_page,
        namespaces={"re": "http://exslt.org/regular-expressions"},
    )

    for link in links:
        # get some identifier for this inside page for caching
        grp = re.match(group_page, link.get("href")).group(1)

        # download inside page, find the matching links
        page = utils.download(
            urllib.parse.urljoin(index_page, link.get("href")),
            "%s/votes/%s/pages/house_%s.html" % (congress, session_year, grp),
            options,
        )

        if not page:
            logging.error(
                "Couldn't download House vote group page (%s), aborting" % grp
            )
            continue

        doc = html.document_fromstring(page)
        votelinks = doc.xpath(
            "//a[re:match(@href, '%s')]" % link_pattern,
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        for votelink in votelinks:
            num = re.match(link_pattern, votelink.get("href")).group(1)
            vote_id = "h" + num + "-" + str(congress) + "." + session_year
            if not should_process(vote_id, options):
                continue
            vote_ids.append(vote_id)

    return utils.uniq(vote_ids)


def vote_ids_for_senate(congress, session_year, options):
    session_num = int(session_year) - utils.get_congress_first_year(int(congress)) + 1

    vote_ids = []

    url = (
        "https://www.senate.gov/legislative/LIS/roll_call_lists/vote_menu_%s_%d.xml"
        % (congress, session_num)
    )

    download_options = options.copy()
    if options.get("fast", False):
        download_options["force"] = True  # Force fresh download when using --fast

    page = utils.download(
        url,
        "%s/votes/%s/pages/senate.xml" % (congress, session_year),
        utils.merge(download_options, {"binary": True}),
    )

    if not page or b"Requested Page Not Found (404)" in page:
        logging.error("Couldn't download Senate vote XML index %s, skipping" % url)
        return None

    dom = etree.fromstring(page)

    # Sanity checks.
    if int(congress) != int(dom.xpath("congress")[0].text):
        logging.error(
            "Senate vote XML returns the wrong Congress: %s"
            % dom.xpath("congress")[0].text
        )
        return None
    if int(session_year) != int(dom.xpath("congress_year")[0].text):
        logging.error(
            "Senate vote XML returns the wrong session: %s"
            % dom.xpath("congress_year")[0].text
        )
        return None

    # Get vote list.
    for vote in dom.xpath("//vote"):
        num = int(vote.xpath("vote_number")[0].text)
        vote_id = "s" + str(num) + "-" + str(congress) + "." + session_year
        if not should_process(vote_id, options):
            continue
        vote_ids.append(vote_id)
    return vote_ids


def should_process(vote_id, options):
    if not options.get("fast", False):
        return True

    # If --fast is used, only download new votes or votes taken in the last
    # three days (when most vote changes and corrections should occur).
    f = vote_info.output_for_vote(vote_id, "json")
    if not os.path.exists(f):
        return True

    v = json.load(open(f))
    now = utils.eastern_time_zone.localize(datetime.datetime.now())
    lookback = options.get("lookback", 3)
    return (now - iso8601.parse_date(v["date"])) < datetime.timedelta(days=lookback)



"""Setup file for using congress as a python package."""

from os import path

import setuptools

# Obtain long_description from README.md
here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="pb-congress-scraper",
    version="0.0.8",
    author="The unitedstates organization on GitHub",
    long_description=long_description,
    description="Public domain data collectors for the work of Congress, "
    "including legislation, amendments, and votes.",
    license="CC0-1.0",
    packages=setuptools.find_packages(),
    install_requires=[
        "beautifulsoup4",
        "cssselect",
        "iso8601",
        "lxml",
        "mechanize",
        "mock",
        "rtyaml",
        "python-dateutil",
        "pytz",
        "pyyaml",
        "scrapelib",
        "xmltodict",
        "packaging",
    ],
    entry_points={
        "console_scripts": ["usc-run=congress.run:main"],
    },
    scripts=[
        "scripts/bills.sh",
        "scripts/statutes.sh",
        "scripts/votes.sh",
        "scripts/voteview.sh",
    ],
)



from congress.tasks import utils
import os
import os.path
import re
from lxml import html, etree
import logging

from congress.tasks import nomination_info


def run(options):
    nomination_id = options.get("nomination_id", None)

    if nomination_id:
        nomination_type, number, congress = utils.split_nomination_id(nomination_id)
        to_fetch = [nomination_id]
    else:
        congress = options.get("congress", utils.current_congress())
        to_fetch = nomination_ids_for(congress, options)
        if not to_fetch:
            if options.get("fast", False):
                logging.warn("No nominations changed.")
            else:
                logging.error(
                    "Error figuring out which nominations to download, aborting."
                )
            return None

        limit = options.get("limit", None)
        if limit:
            to_fetch = to_fetch[: int(limit)]

    logging.warn(
        "Going to fetch %i nominations from congress #%s" % (len(to_fetch), congress)
    )

    saved_nominations = utils.process_set(
        to_fetch, nomination_info.fetch_nomination, options
    )


# page through listings for bills of a particular congress


def nomination_ids_for(congress, options={}):
    nomination_ids = []

    page = page_for(congress, options)
    if not page:
        logging.error("Couldn't download page for %d congress" % congress)
        return None

    # extract matching links
    doc = html.document_fromstring(page)
    raw_nomination_ids = doc.xpath('//div[@id="content"]/p[2]/a/text()')
    nomination_ids = []

    for raw_id in raw_nomination_ids:
        pieces = raw_id.split(" ")

        # ignore these
        if raw_id in ["PDF", "Text", "split into two or more parts"]:
            pass
        elif len(pieces) < 2:
            logging.error("Bad nomination ID detected: %s" % raw_id)
            return None
        else:
            nomination_ids.append(pieces[1])

    return utils.uniq(nomination_ids)


def page_cache_for(congress):
    return "%s/nominations/pages/search.html" % congress


# unlike bills.py, we're going to fetch the page instead of producing the URL,
# since a POST is required.


def page_for(congress, options):
    congress = int(congress)
    postdata = {
        "database": "nominations",
        "MaxDocs": "5000",
        "submit": "SEARCH",
        "querytype": "phrase",
        "query": "",
        "Stemming": "No",
        "congress": "%d" % congress,
        "CIVcategory": "on",
        "LSTcategory": "on",
        "committee": "",
        "LBDateSel": "FLD606",
        "EBSDate": "",
        "EBEDate": "",
        "sort": "sh_docid_rc",
    }

    post_options = {"postdata": postdata}
    post_options.update(options)

    # unused: never cache search listing
    cache = page_cache_for(congress)

    page = utils.download("http://thomas.loc.gov/cgi-bin/thomas", None, post_options)
    return page
